import warnings

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from loguru import logger
from sklearn.base import clone, is_classifier, is_regressor
from sklearn.model_selection import check_cv
from sklearn.model_selection._search import BaseSearchCV

from probatus.utils import (
    BaseFitComputePlotClass,
    assure_pandas_series,
    calculate_shap_importance,
    preprocess_data,
    preprocess_labels,
    get_single_scorer,
    shap_calc,
)


class ShapRFECV(BaseFitComputePlotClass):
    """
    This class performs Backwards Recursive Feature Elimination, using SHAP feature importance.

    At each round, for a
        given feature set, starting from all available features, the following steps are applied:

    1. (Optional) Tune the hyperparameters of the model using sklearn compatible search CV e.g.
        [GridSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LassoCV.html),
        [RandomizedSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.RandomizedSearchCV.html?highlight=randomized#sklearn.model_selection.RandomizedSearchCV), or
        [BayesSearchCV](https://scikit-optimize.github.io/stable/modules/generated/skopt.BayesSearchCV.html),
    2. Apply Cross-validation (CV) to estimate the SHAP feature importance on the provided dataset. In each CV
        iteration, the model is fitted on the train folds, and applied on the validation fold to estimate
        SHAP feature importance.
    3. Remove `step` lowest SHAP importance features from the dataset.

    At the end of the process, the user can plot the performance of the model for each iteration, and select the
        optimal number of features and the features set.

    The functionality is
        similar to [RFECV](https://scikit-learn.org/stable/modules/generated/sklearn.feature_selection.RFECV.html).
        The main difference is removing the lowest importance features based on SHAP features importance. It also
        supports the use of sklearn compatible search CV for hyperparameter optimization e.g.
        [GridSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LassoCV.html),
        [RandomizedSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.RandomizedSearchCV.html?highlight=randomized#sklearn.model_selection.RandomizedSearchCV), or
        [BayesSearchCV](https://scikit-optimize.github.io/stable/modules/generated/skopt.BayesSearchCV.html), which
        needs to be passed as the `model`. Thanks to this you can perform hyperparameter optimization at each step of
        the feature elimination. Lastly, it supports categorical features (object and category dtype) and missing values
        in the data, as long as the model supports them.

    We recommend using [LGBMClassifier](https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMClassifier.html),
        because by default it handles missing values and categorical features. In case of other models, make sure to
        handle these issues for your dataset and consider impact it might have on features importance.


    Example:
    ```python
    import numpy as np
    import pandas as pd
    from probatus.feature_elimination import ShapRFECV
    from sklearn.datasets import make_classification
    from sklearn.model_selection import train_test_split
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.model_selection import RandomizedSearchCV

    feature_names = [
        'f1', 'f2', 'f3', 'f4', 'f5', 'f6', 'f7',
        'f8', 'f9', 'f10', 'f11', 'f12', 'f13',
        'f14', 'f15', 'f16', 'f17', 'f18', 'f19', 'f20']

    # Prepare two samples
    X, y = make_classification(n_samples=200, class_sep=0.05, n_informative=6, n_features=20,
                               random_state=0, n_redundant=10, n_clusters_per_class=1)
    X = pd.DataFrame(X, columns=feature_names)


    # Prepare model and parameter search space
    model = RandomForestClassifier(max_depth=5, class_weight='balanced')

    param_grid = {
        'n_estimators': [5, 7, 10],
        'min_samples_leaf': [3, 5, 7, 10],
    }
    search = RandomizedSearchCV(model, param_grid)


    # Run feature elimination
    shap_elimination = ShapRFECV(
        model=search, step=0.2, cv=10, scoring='roc_auc', n_jobs=3)
    report = shap_elimination.fit_compute(X, y)

    # Make plots
    performance_plot = shap_elimination.plot()

    # Get final feature set
    final_features_set = shap_elimination.get_reduced_features_set(num_features=3)
    ```
    <img src="../img/shaprfecv.png" width="500" />

    """  # noqa

    def __init__(
        self,
        model,
        step=1,
        min_features_to_select=1,
        cv=None,
        scoring="roc_auc",
        n_jobs=-1,
        verbose=0,
        random_state=None,
    ):
        """
        This method initializes the class.

        Args:
            model (classifier or regressor, sklearn compatible search CV e.g. GridSearchCV, RandomizedSearchCV or BayesSearchCV):
                A model that will be optimized and trained at each round of feature elimination. The recommended model
                is [LGBMClassifier](https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMClassifier.html),
                because it by default handles the missing values and categorical variables. This parameter also supports
                any hyperparameter search schema that is consistent with the sklearn API e.g.
                [GridSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.GridSearchCV.html),
                [RandomizedSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.RandomizedSearchCV.html)
                or [BayesSearchCV](https://scikit-optimize.github.io/stable/modules/generated/skopt.BayesSearchCV.html#skopt.BayesSearchCV).

            step (int or float, optional):
                Number of lowest importance features removed each round. If it is an int, then each round such a number of
                features are discarded. If float, such a percentage of remaining features (rounded down) is removed each
                iteration. It is recommended to use float, since it is faster for a large number of features, and slows
                down and becomes more precise with fewer features. Note: the last round may remove fewer features in
                order to reach min_features_to_select.
                If columns_to_keep parameter is specified in the fit method, step is the number of features to remove after
                keeping those columns.

            min_features_to_select (int, optional):
                Minimum number of features to be kept. This is a stopping criterion of the feature elimination. By
                default the process stops when one feature is left. If columns_to_keep is specified in the fit method,
                it may override this parameter to the maximum between length of columns_to_keep the two.

            cv (int, cross-validation generator or an iterable, optional):
                Determines the cross-validation splitting strategy. Compatible with sklearn
                [cv parameter](https://scikit-learn.org/stable/modules/generated/sklearn.feature_selection.RFECV.html).
                If None, then cv of 5 is used.

            scoring (string or probatus.utils.Scorer, optional):
                Metric for which the model performance is calculated. It can be either a metric name aligned with predefined
                [classification scorers names in sklearn](https://scikit-learn.org/stable/modules/model_evaluation.html).

            n_jobs (int, optional):
                Number of cores to run in parallel while fitting across folds. None means 1 unless in a
                `joblib.parallel_backend` context. -1 means using all processors.

            verbose (int, optional):
                Controls verbosity of the output:

                - 0 - neither prints nor warnings are shown
                - 1 - only most important warnings
                - 2 - shows all prints and all warnings.

            random_state (int, optional):
                Random state set at each round of feature elimination. If it is None, the results will not be
                reproducible and in random search at each iteration a different hyperparameters might be tested. For
                reproducible results set it to an integer.
        """  # noqa
        self.model = model
        self.search_model = isinstance(model, BaseSearchCV)
        self.step = self._validate_step(step)
        self.min_features_to_select = self._validate_min_features(min_features_to_select)
        self.cv = cv
        self.scorer = get_single_scorer(scoring)
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.random_state = random_state
        self.report_df = pd.DataFrame()

    def compute(self):
        """
        Checks if fit() method has been run.

        and computes the DataFrame with results of feature elimination for each round.

        Returns:
            (pd.DataFrame):
                DataFrame with results of feature elimination for each round.
        """
        self._check_if_fitted()

        return self.report_df

    def fit_compute(
        self,
        X,
        y,
        sample_weight=None,
        columns_to_keep=None,
        column_names=None,
        shap_variance_penalty_factor=None,
        **shap_kwargs,
    ):
        """
        Fits the object with the provided data.

        The algorithm starts with the entire dataset, and then sequentially
            eliminates features. If sklearn compatible search CV is passed as model e.g.
            [GridSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.GridSearchCV.html),
            [RandomizedSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.RandomizedSearchCV.html)
            or [BayesSearchCV](https://scikit-optimize.github.io/stable/modules/generated/skopt.BayesSearchCV.html),
            the hyperparameter optimization is applied at each step of the elimination.
            Then, the SHAP feature importance is calculated using Cross-Validation,
            and `step` lowest importance features are removed. At the end, the
            report containing results from each iteration is computed and returned to the user.

        Args:
            X (pd.DataFrame):
                Provided dataset.

            y (pd.Series):
                Labels for X.

            sample_weight (pd.Series, np.ndarray, list, optional):
                array-like of shape (n_samples,) - only use if the model you're using supports
                sample weighting (check the corresponding scikit-learn documentation).
                Array of weights that are assigned to individual samples.
                Note that they're only used for fitting of  the model, not during evaluation of metrics.
                If not provided, then each sample is given unit weight.

            columns_to_keep (list of str, optional):
                List of columns to keep. If given, these columns will not be eliminated.

            column_names (list of str, optional):
                List of feature names of the provided samples. If provided it will be used to overwrite the existing
                feature names. If not provided the existing feature names are used or default feature names are
                generated.

            shap_variance_penalty_factor (int or float, optional):
                Apply aggregation penalty when computing average of shap values for a given feature.
                Results in a preference for features that have smaller standard deviation of shap
                values (more coherent shap importance). Recommend value 0.5 - 1.0.
                Formula: penalized_shap_mean = (mean_shap - (std_shap * shap_variance_penalty_factor))

            **shap_kwargs:
                keyword arguments passed to
                [shap.Explainer](https://shap.readthedocs.io/en/latest/generated/shap.Explainer.html#shap.Explainer).
                It also enables `approximate` and `check_additivity` parameters, passed while calculating SHAP values.
                The `approximate=True` causes less accurate, but faster SHAP values calculation, while
                `check_additivity=False` disables the additivity check inside SHAP.

        Returns:
            (pd.DataFrame):
                DataFrame containing results of feature elimination from each iteration.
        """

        self.fit(
            X,
            y,
            sample_weight=sample_weight,
            columns_to_keep=columns_to_keep,
            column_names=column_names,
            shap_variance_penalty_factor=shap_variance_penalty_factor,
            **shap_kwargs,
        )
        return self.compute()

    def fit(
        self,
        X,
        y,
        sample_weight=None,
        columns_to_keep=None,
        column_names=None,
        groups=None,
        shap_variance_penalty_factor=None,
        **shap_kwargs,
    ):
        """
        Fits the object with the provided data.

        The algorithm starts with the entire dataset, and then sequentially
            eliminates features. If sklearn compatible search CV is passed as model e.g.
            [GridSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.GridSearchCV.html),
            [RandomizedSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.RandomizedSearchCV.html)
            or [BayesSearchCV](https://scikit-optimize.github.io/stable/modules/generated/skopt.BayesSearchCV.html),
            the hyperparameter optimization is applied at each step of the elimination.
            Then, the SHAP feature importance is calculated using Cross-Validation,
            and `step` lowest importance features are removed.

        Args:
            X (pd.DataFrame):
                Provided dataset.

            y (pd.Series):
                Labels for X.

            sample_weight (pd.Series, np.ndarray, list, optional):
                array-like of shape (n_samples,) - only use if the model you're using supports
                sample weighting (check the corresponding scikit-learn documentation).
                Array of weights that are assigned to individual samples.
                Note that they're only used for fitting of  the model, not during evaluation of metrics.
                If not provided, then each sample is given unit weight.

            columns_to_keep (list of str, optional):
                List of column names to keep. If given,
                these columns will not be eliminated by the feature elimination process.
                However, these feature will used for the calculation of the SHAP values.

            column_names (list of str, optional):
                List of feature names of the provided samples. If provided it will be used to overwrite the existing
                feature names. If not provided the existing feature names are used or default feature names are
                generated.

            groups (pd.Series, np.ndarray, list, optional):
                array-like of shape (n_samples,)
                Group labels for the samples used while splitting the dataset into train/test set.
                Only used in conjunction with a "Group" `cv` instance.
                (e.g. `sklearn.model_selection.GroupKFold`).

            shap_variance_penalty_factor (int or float, optional):
                Apply aggregation penalty when computing average of shap values for a given feature.
                Results in a preference for features that have smaller standard deviation of shap
                values (more coherent shap importance). Recommend value 0.5 - 1.0.
                Formula: penalized_shap_mean = (mean_shap - (std_shap * shap_variance_penalty_factor))

            **shap_kwargs:
                keyword arguments passed to
                [shap.Explainer](https://shap.readthedocs.io/en/latest/generated/shap.Explainer.html#shap.Explainer).
                It also enables `approximate` and `check_additivity` parameters, passed while calculating SHAP values.
                The `approximate=True` causes less accurate, but faster SHAP values calculation, while
                `check_additivity=False` disables the additivity check inside SHAP.

        Returns:
            (ShapRFECV): Fitted object.
        """
        # Initialise len_columns_to_keep based on columns_to_keep content validation
        len_columns_to_keep = 0
        if columns_to_keep:
            if not all(isinstance(x, str) for x in columns_to_keep):
                raise ValueError("All elements in columns_to_keep must be strings.")
            len_columns_to_keep = len(columns_to_keep)

        # Validate matching column names, if both columns_to_keep and column_names are provided
        if column_names and not all(x in column_names for x in list(X.columns)):
            raise ValueError("Column names in columns_to_keep and column_names do not match.")

        # Validate total number of columns to select against the total number of columns
        if (
            column_names
            and columns_to_keep
            and (self.min_features_to_select + len_columns_to_keep) > len(self.column_names)
        ):
            raise ValueError("Minimum features to select plus columns_to_keep exceeds total number of features.")

        # Check shap_variance_penalty_factor has acceptable value
        if isinstance(shap_variance_penalty_factor, (float, int)) and shap_variance_penalty_factor >= 0:
            _shap_variance_penalty_factor = shap_variance_penalty_factor
        else:
            if shap_variance_penalty_factor is not None:
                warnings.warn(
                    "shap_variance_penalty_factor must be None, int or float. Setting shap_variance_penalty_factor = 0"
                )
            _shap_variance_penalty_factor = 0

        self.X, self.column_names = preprocess_data(X, X_name="X", column_names=column_names, verbose=self.verbose)
        self.y = preprocess_labels(y, y_name="y", index=self.X.index, verbose=self.verbose)
        if sample_weight is not None:
            if self.verbose > 0:
                warnings.warn(
                    "sample_weight is passed only to the fit method of the model, not the evaluation metrics."
                )
            sample_weight = assure_pandas_series(sample_weight, index=self.X.index)
        self.cv = check_cv(self.cv, self.y, classifier=is_classifier(self.model))

        remaining_features = current_features_set = self.column_names
        round_number = 0

        # Stop when stopping criteria is met.
        stopping_criteria = np.max([self.min_features_to_select, len_columns_to_keep])

        # Setting up the min_features_to_select parameter.
        if columns_to_keep is not None:
            self.min_features_to_select = 0
            # Ensures that, if columns_to_keep is provided, the last features remaining are only the columns_to_keep.
            if self.verbose > 1:
                warnings.warn(f"Minimum features to select : {stopping_criteria}")

        while len(current_features_set) > stopping_criteria:
            round_number += 1

            # Get current dataset info
            current_features_set = remaining_features
            remaining_removeable_features = list(dict.fromkeys(current_features_set + (columns_to_keep or [])))

            # Current dataset
            current_X = self.X[remaining_removeable_features]

            # Optimize parameters
            if self.search_model:
                current_search_model = clone(self.model).fit(current_X, self.y)
                current_model = current_search_model.estimator.set_params(**current_search_model.best_params_)
            else:
                current_model = clone(self.model)

            # Perform CV to estimate feature importance with SHAP
            results_per_fold = Parallel(n_jobs=self.n_jobs)(
                delayed(self._get_feature_shap_values_per_fold)(
                    X=current_X,
                    y=self.y,
                    model=current_model,
                    train_index=train_index,
                    val_index=val_index,
                    sample_weight=sample_weight,
                    **shap_kwargs,
                )
                for train_index, val_index in self.cv.split(current_X, self.y, groups)
            )

            if self.y.nunique() == 2 or is_regressor(current_model):
                shap_values = np.concatenate([current_result[0] for current_result in results_per_fold], axis=0)
            else:  # multi-class case
                shap_values = np.concatenate([current_result[0] for current_result in results_per_fold], axis=1)

            scores_train = [current_result[1] for current_result in results_per_fold]
            scores_val = [current_result[2] for current_result in results_per_fold]

            # Calculate the shap features with remaining features and features to keep.
            shap_importance_df = calculate_shap_importance(
                shap_values, remaining_removeable_features, shap_variance_penalty_factor=_shap_variance_penalty_factor
            )

            # Determine which features to keep and which to remove.
            remaining_features, features_to_remove = self._filter_and_identify_features_based_on_importance(
                shap_importance_df, columns_to_keep, current_features_set
            )

            # Report results
            self._report_current_results(
                round_number=round_number,
                current_features_set=current_features_set,
                features_to_remove=features_to_remove,
                train_metric_mean=np.mean(scores_train),
                train_metric_std=np.std(scores_train),
                val_metric_mean=np.mean(scores_val),
                val_metric_std=np.std(scores_val),
            )
            if self.verbose > 1:
                logger.info(
                    f"Round: {round_number}, Current number of features: {len(current_features_set)}, "
                    f'Current performance: Train {self.report_df.loc[round_number]["train_metric_mean"]} '
                    f'+/- {self.report_df.loc[round_number]["train_metric_std"]}, CV Validation '
                    f'{self.report_df.loc[round_number]["val_metric_mean"]} '
                    f'+/- {self.report_df.loc[round_number]["val_metric_std"]}. \n'
                    f"Features left: {remaining_features}. "
                    f"Removed features at the end of the round: {features_to_remove}"
                )
        self.fitted = True
        return self

    def plot(self, show=True, **figure_kwargs):
        """
        Generates plot of the model performance for each iteration of feature elimination.

        Args:
            show (bool, optional):
                If True, the plots are showed to the user, otherwise they are not shown. Not showing plot can be useful,
                when you want to edit the returned figure, before showing it.

            **figure_kwargs:
                Keyword arguments that are passed to the plt.figure, at its initialization.

        Returns:
            (plt.figure):
                Figure containing the performance plot.
        """
        # Data preparation
        num_features = self.report_df["num_features"]
        train_mean = self.report_df["train_metric_mean"]
        train_std = self.report_df["train_metric_std"]
        val_mean = self.report_df["val_metric_mean"]
        val_std = self.report_df["val_metric_std"]
        x_ticks = list(reversed(num_features.tolist()))

        # Plotting
        fig, ax = plt.subplots(**figure_kwargs)

        # Training performance
        ax.plot(num_features, train_mean, label="Train Score")
        ax.fill_between(num_features, train_mean - train_std, train_mean + train_std, alpha=0.3)

        # Validation performance
        ax.plot(num_features, val_mean, label="Validation Score")
        ax.fill_between(num_features, val_mean - val_std, val_mean + val_std, alpha=0.3)

        # Labels and title
        ax.set_xlabel("Number of features")
        ax.set_ylabel(f"Performance {self.scorer.metric_name}")
        ax.set_title("Backwards Feature Elimination using SHAP & CV")
        ax.legend(loc="lower left")
        ax.invert_xaxis()
        ax.set_xticks(x_ticks)

        # Display or close plot
        if show:
            plt.show()
        else:
            plt.close(fig)

        return fig

    @staticmethod
    def _validate_step(step):
        if not isinstance(step, (int, float)) or step <= 0:
            raise ValueError(f"Invalid step value: {step}. Must be a positive int or float.")
        return step

    @staticmethod
    def _validate_min_features(min_features):
        if not isinstance(min_features, int) or min_features <= 0:
            raise ValueError(f"Invalid min_features_to_select value: {min_features}. Must be a positive int.")
        return min_features

    @staticmethod
    def _calculate_number_of_features_to_remove(
        current_num_of_features,
        num_features_to_remove,
        min_num_features_to_keep,
    ):
        """
        Calculates the number of features to be removed.

        Makes sure that after removal at least
            min_num_features_to_keep are kept

        Args:
            current_num_of_features (int):
                Current number of features in the data.

            num_features_to_remove (int):
                Number of features to be removed at this stage.

            min_num_features_to_keep (int):
                Minimum number of features to be left after removal.

        Returns:
            (int):
                Number of features to be removed.
        """
        # Calculate maximum nr of features that can be removed without dropping below
        # `min_num_features_to_keep`.
        nr_of_max_allowed_feature_removed = current_num_of_features - min_num_features_to_keep

        # Return smallest between `nr_of_max_allowed_feature_removed` and `num_features_to_remove`
        return min(num_features_to_remove, nr_of_max_allowed_feature_removed)

    def _get_current_features_to_remove(self, shap_importance_df, columns_to_keep=None):
        """
        Implements the logic used to determine which features to remove.

        If step is a positive integer,
            at each round step lowest SHAP importance features are selected. If it is a float, such percentage
            of remaining features (rounded up) is removed each iteration. It is recommended to use float, since it is
            faster for a large set of features, and slows down and becomes more precise with fewer features.

        Args:
            shap_importance_df (pd.DataFrame):
                DataFrame presenting SHAP importance of remaining features.

            columns_to_keep Optional(list)L
                A list of features that are kept.

        Returns:
            (list):
                List of features to be removed at a given round.
        """
        # Bounding the variable.
        num_features_to_remove = 0

        # If columns_to_keep is not None, exclude those columns and
        # calculate features to remove.
        if columns_to_keep is not None:
            mask = shap_importance_df.index.isin(columns_to_keep)
            shap_importance_df = shap_importance_df[~mask]

        # If the step is an int remove n features.
        if isinstance(self.step, int):
            num_features_to_remove = self._calculate_number_of_features_to_remove(
                current_num_of_features=shap_importance_df.shape[0],
                num_features_to_remove=self.step,
                min_num_features_to_keep=self.min_features_to_select,
            )
        # If the step is a float remove n * number features that are left, rounded down
        elif isinstance(self.step, float):
            current_step = int(np.floor(shap_importance_df.shape[0] * self.step))
            # The step after rounding down should be at least 1
            if current_step < 1:
                current_step = 1

            num_features_to_remove = self._calculate_number_of_features_to_remove(
                current_num_of_features=shap_importance_df.shape[0],
                num_features_to_remove=current_step,
                min_num_features_to_keep=self.min_features_to_select,
            )

        if num_features_to_remove == 0:
            return []
        else:
            return shap_importance_df.iloc[-num_features_to_remove:].index.tolist()

    def _report_current_results(
        self,
        round_number,
        current_features_set,
        features_to_remove,
        train_metric_mean,
        train_metric_std,
        val_metric_mean,
        val_metric_std,
    ):
        """
        This function adds the results from a current iteration to the report.

        Args:
            round_number (int):
                Current number of the round.

            current_features_set (list of str):
                Current list of features.

            features_to_remove (list of str):
                List of features to be removed at the end of this iteration.

            train_metric_mean (float or int):
                Mean scoring metric measured on train set during CV.

            train_metric_std (float or int):
                Std scoring metric measured on train set during CV.

            val_metric_mean (float or int):
                Mean scoring metric measured on validation set during CV.

            val_metric_std (float or int):
                Std scoring metric measured on validation set during CV.
        """
        current_results = {
            "num_features": len(current_features_set),
            "features_set": [current_features_set],
            "eliminated_features": [features_to_remove],
            "train_metric_mean": train_metric_mean,
            "train_metric_std": train_metric_std,
            "val_metric_mean": val_metric_mean,
            "val_metric_std": val_metric_std,
        }

        if self.report_df.empty:
            self.report_df = pd.DataFrame(current_results, index=[round_number])
        else:
            new_row = pd.DataFrame(current_results, index=[round_number])
            # Append new_row to self.report_df more efficiently
            self.report_df = pd.concat([self.report_df, new_row])

    def _get_feature_shap_values_per_fold(
        self,
        X,
        y,
        model,
        train_index,
        val_index,
        sample_weight=None,
        **shap_kwargs,
    ):
        """
        This function calculates the shap values on validation set, and Train and Val score.

        Args:
            X (pd.DataFrame):
                Dataset used in CV.

            y (pd.Series):
                Labels for X.

            model (classifier or regressor):
                Model to be fitted on the train folds.

            train_index (np.array):
                Positions of train folds samples.

            val_index (np.array):
                Positions of validation fold samples.

            sample_weight (pd.Series, np.ndarray, list, optional):
                array-like of shape (n_samples,) - only use if the model you're using supports
                sample weighting (check the corresponding scikit-learn documentation).
                Array of weights that are assigned to individual samples.
                Note that they're only used for fitting of  the model, not during evaluation of metrics.
                If not provided, then each sample is given unit weight.

            **shap_kwargs:
                keyword arguments passed to
                [shap.Explainer](https://shap.readthedocs.io/en/latest/generated/shap.Explainer.html#shap.Explainer).
                It also enables `approximate` and `check_additivity` parameters, passed while calculating SHAP values.
                The `approximate=True` causes less accurate, but faster SHAP values calculation, while
                `check_additivity=False` disables the additivity check inside SHAP.
        Returns:
            (np.array, float, float):
                Tuple with the results: Shap Values on validation fold, train score, validation score.
        """
        X_train, X_val = X.iloc[train_index, :], X.iloc[val_index, :]
        y_train, y_val = y.iloc[train_index], y.iloc[val_index]

        if sample_weight is not None:
            model = model.fit(X_train, y_train, sample_weight=sample_weight.iloc[train_index])
        else:
            model = model.fit(X_train, y_train)

        # Score the model
        score_train = self.scorer.score(model, X_train, y_train)
        score_val = self.scorer.score(model, X_val, y_val)

        # Compute SHAP values
        shap_values = shap_calc(model, X_val, verbose=self.verbose, random_state=self.random_state, **shap_kwargs)
        return shap_values, score_train, score_val

    def _filter_and_identify_features_based_on_importance(
        self, shap_importance_df, columns_to_keep, current_features_set
    ):
        """
        Filters out features to be removed from the current feature set based on SHAP importance,
        while maintaining the original order of the features.

        Args:
            shap_importance_df (pd.DataFrame):
                A DataFrame containing the SHAP importance of the features.

            columns_to_keep (list):
                A list of column names that should not be removed, regardless of their
                SHAP importance.

            current_features_set (list):
                The current list of features from which features identified as
                less important will be removed. This list's order is maintained in the
                returned list of remaining features.

        Returns:
            remaining_features, features_to_remove (list, list): The features to keep & those that are removed.
        """
        # Get features to remove based on SHAP importance and columns to keep
        features_to_remove = self._get_current_features_to_remove(shap_importance_df, columns_to_keep=columns_to_keep)

        # Convert features_to_remove to a set for O(1) lookup times
        features_to_remove_set = set(features_to_remove)

        # Filter out the features to remove, maintaining the original order of current_features_set
        remaining_features = [feature for feature in current_features_set if feature not in features_to_remove_set]

        return remaining_features, features_to_remove

    def get_reduced_features_set(self, num_features, standard_error_threshold=1.0, return_type="feature_names"):
        """
        Gets the features set after the feature elimination process, for a given number of features.

        Args:
            num_features (int or str):
                If int: Number of features in the reduced features set.
                If str: One of the following automatic num feature selection methods supported:
                    1. best: strictly selects the num_features with the highest model score.
                    2. best_coherent: For iterations that are within standard_error_threshold of the highest
                    score, select the iteration with the lowest standard deviation of model score.
                    3. best_parsimonious: For iterations that are within standard_error_threshold of the
                    highest score, select the iteration with the fewest features.

            standard_error_threshold (float):
                If num_features is 'best_coherent' or 'best_parsimonious', this parameter is used.

            return_type:
                Accepts possible values of 'feature_names', 'support' or 'ranking'. These are defined as:
                    1. feature_names: returns column names
                    2. support: returns boolean mask
                    3. ranking: returns numeric ranking of features

        Returns:
            (list of str):
                Reduced features set.
        """
        self._check_if_fitted()

        # Determine the best number of features based on the method specified
        if isinstance(num_features, str):
            num_features = self._get_best_num_features(
                best_method=num_features, standard_error_threshold=standard_error_threshold
            )
        elif not isinstance(num_features, int):
            ValueError(
                "Parameter num_features can be of type int, or of type str with "
                "possible values of 'best', 'best_coherent' or 'best_parsimonious'"
            )

        # Get feature names for the determined number of features
        feature_names_selected = self._get_feature_names(num_features)

        # Return based on the requested return type
        if return_type == "feature_names":
            return feature_names_selected
        elif return_type == "support":
            return self._get_feature_support(feature_names_selected)
        elif return_type == "ranking":
            return self._get_feature_ranking()
        else:
            raise ValueError("Invalid return_type. Must be 'feature_names', 'support', or 'ranking'.")

    def _get_best_num_features(self, best_method, standard_error_threshold=1.0):
        """
        Helper function to identify the best number of features to select as per some automatic
        feature selection strategy. Strategies supported are:
            1. best: strictly selects the num_features with the highest model score.
            2. best_coherent: For iterations that are within standard_error_threshold of the highest
            score, select the iteration with the lowest standard deviation of model score.
            3. best_parsimonious: For iterations that are within standard_error_threshold of the
            highest score, select the iteration with the fewest features.

        Args:
            best_method (str):
                Automatic best feature selection strategy. One of "best", "best_coherent" or
                "best_parsimonious".

            standard_error_threshold (float):
                Parameter used if best_method is 'best_coherent' or 'best_parsimonious'.
                Numeric value greater than zero.

        Returns:
            (int)
                num_features as per automatic feature selection strategy selected.
        """
        self._check_if_fitted()

        if not isinstance(standard_error_threshold, (float, int)) or standard_error_threshold < 0:
            raise ValueError("Parameter standard_error_threshold must be a non-negative int or float.")

        # Perform copy after ValueError check.
        shap_report = self.report_df.copy()

        if best_method == "best":
            # Strictly selects the number of features with the highest model score
            best_score_index = shap_report["val_metric_mean"].idxmax()
            best_num_features = shap_report.loc[best_score_index, "num_features"]

        elif best_method == "best_coherent":
            # Selects within a threshold but prioritizes lower standard deviation
            highest_score = shap_report["val_metric_mean"].max()
            within_threshold = shap_report[shap_report["val_metric_mean"] >= highest_score - standard_error_threshold]
            lowest_std_index = within_threshold["val_metric_std"].idxmin()
            best_num_features = within_threshold.loc[lowest_std_index, "num_features"]

        elif best_method == "best_parsimonious":
            # Selects the fewest number of features within the threshold of the highest score
            highest_score = shap_report["val_metric_mean"].max()
            within_threshold = shap_report[shap_report["val_metric_mean"] >= highest_score - standard_error_threshold]
            fewest_features_index = within_threshold["num_features"].idxmin()
            best_num_features = within_threshold.loc[fewest_features_index, "num_features"]

        else:
            raise ValueError(
                "The parameter 'best_method' must be one of 'best', 'best_coherent', or 'best_parsimonious'."
            )

        # Log shap_report for users who want to inspect / debug
        if self.verbose > 1:
            logger.info(shap_report)

        return best_num_features

    def _get_feature_names(self, num_features):
        """
        Helper function that takes num_features and returns the associated list of column/feature names.

        Args:
            num_features (int):
                Represents the top N features to get the column names for.

        Returns:
            (list of feature names)
                List of the names of the features representing top num_features
        """
        self._check_if_fitted()

        # Direct lookup for the row with the desired number of features
        matching_rows = self.report_df[self.report_df.num_features == num_features]

        if matching_rows.empty:
            valid_nums = ", ".join([str(n) for n in sorted(self.report_df.num_features.unique())])
            raise ValueError(
                f"The provided number of features has not been achieved at any stage of the process. "
                f"You can select one of the following: {valid_nums}"
            )

        # Assuming 'features_set' contains the list of feature names for the row
        return matching_rows.iloc[0]["features_set"]

        # Assuming 'features_set' contains the list of feature names for the row
        return matching_rows.iloc[0]["features_set"]

    @staticmethod
    def _get_feature_support(self, feature_names_selected):
        """
        Helper function that takes feature_names_selected and returns a boolean mask representing the columns
        that were selected by the RFECV method.

        Args:
            feature_names_selected (list):
                Represents the top N features to get the column names for.

        Returns:
            (list of bools)
                Boolean mask representing the features selected.
        """
        support = [True if col in feature_names_selected else False for col in self.column_names]

        return support

    def _get_feature_ranking(self):
        """
        Returns the feature ranking, such that ranking_[i] corresponds to the ranking position
        of the i-th feature. Selected (i.e., estimated best) features are assigned rank 1.

        Returns:
            (list of bools)
                Boolean mask representing the features selected.
        """
        flipped_report_df = self.report_df.iloc[::-1]

        # Some features are not eliminated. All have importance of zero (highest importance)
        features_not_eliminated = flipped_report_df["features_set"].iloc[0]
        features_not_eliminated_dict = {v: 0 for v in features_not_eliminated}

        # Eliminated features are ranked by shap importance
        features_eliminated = np.concatenate(flipped_report_df["eliminated_features"].to_numpy())
        features_eliminated_dict = {int(v): k + 1 for (k, v) in enumerate(features_eliminated)}

        # Combine dicts with rank info
        features_eliminated_dict.update(features_not_eliminated_dict)

        # Get ranking per the order of columns
        ranking = [features_eliminated_dict[col] for col in self.column_names]

        return ranking



def assure_list_of_strings(variable, variable_name):
    """
    Make sure object is a list of strings.
    """
    if isinstance(variable, list):
        return variable
    elif isinstance(variable, str):
        return [variable]
    else:
        raise (ValueError("{} needs to be either a string or list of strings.").format(variable_name))



import warnings

from probatus.utils import (
    shap_calc,
)
from probatus.feature_elimination import ShapRFECV


class EarlyStoppingShapRFECV(ShapRFECV):
    """
    This class performs Backwards Recursive Feature Elimination, using SHAP feature importance.

    This is a child of ShapRFECV which allows early stopping of the training step, this class is compatible with
        LightGBM, XGBoost and CatBoost models. If you are not using early stopping, you should use the parent class,
        ShapRFECV, instead of EarlyStoppingShapRFECV.

    [Early stopping](https://en.wikipedia.org/wiki/Early_stopping) is a type of
        regularization technique in which the model is trained until the scoring metric, measured on a validation set,
        stops improving after a number of early_stopping_rounds. In boosted tree models, this technique can increase
        the training speed, by skipping the training of trees that do not improve the scoring metric any further,
        which is particularly useful when the training dataset is large.

    Note that if the regressor or classifier is a hyperparameter search model is used, the early stopping parameter is passed only
        to the fit method of the model duiring the Shapley values estimation step, and not for the hyperparameter
        search step.
        Early stopping can be seen as a type of regularization of the optimal number of trees. Therefore you can use
        it directly with a LightGBM or XGBoost model, as an alternative to a hyperparameter search model.

    At each round, for a
        given feature set, starting from all available features, the following steps are applied:

    1. (Optional) Tune the hyperparameters of the model using sklearn compatible search CV e.g.
        [GridSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LassoCV.html),
        [RandomizedSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.RandomizedSearchCV.html?highlight=randomized#sklearn.model_selection.RandomizedSearchCV), or
        [BayesSearchCV](https://scikit-optimize.github.io/stable/modules/generated/skopt.BayesSearchCV.html).
        Note that during this step the model does not use early stopping.
    2. Apply Cross-validation (CV) to estimate the SHAP feature importance on the provided dataset. In each CV
        iteration, the model is fitted on the train folds, and applied on the validation fold to estimate
        SHAP feature importance. The model is trained until the scoring metric eval_metric, measured on the
        validation fold, stops improving after a number of early_stopping_rounds.
    3. Remove `step` lowest SHAP importance features from the dataset.

    At the end of the process, the user can plot the performance of the model for each iteration, and select the
        optimal number of features and the features set.

    We recommend using [LGBMClassifier](https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMClassifier.html),
        because by default it handles missing values and categorical features. In case of other models, make sure to
        handle these issues for your dataset and consider impact it might have on features importance.


    Example:
    ```python
    from lightgbm import LGBMClassifier
    import pandas as pd
    from probatus.feature_elimination import EarlyStoppingShapRFECV
    from sklearn.datasets import make_classification

    feature_names = [
        'f1', 'f2', 'f3', 'f4', 'f5', 'f6', 'f7',
        'f8', 'f9', 'f10', 'f11', 'f12', 'f13',
        'f14', 'f15', 'f16', 'f17', 'f18', 'f19', 'f20']

    # Prepare two samples
    X, y = make_classification(n_samples=200, class_sep=0.05, n_informative=6, n_features=20,
                               random_state=0, n_redundant=10, n_clusters_per_class=1)
    X = pd.DataFrame(X, columns=feature_names)

    # Prepare model
    model = LGBMClassifier(n_estimators=200, max_depth=3)

    # Run feature elimination
    shap_elimination = EarlyStoppingShapRFECV(
        model=model, step=0.2, cv=10, scoring='roc_auc', early_stopping_rounds=10, n_jobs=3)
    report = shap_elimination.fit_compute(X, y)

    # Make plots
    performance_plot = shap_elimination.plot()

    # Get final feature set
    final_features_set = shap_elimination.get_reduced_features_set(num_features=3)
    ```
    <img src="../img/earlystoppingshaprfecv.png" width="500" />

    """  # noqa

    def __init__(
        self,
        model,
        step=1,
        min_features_to_select=1,
        cv=None,
        scoring="roc_auc",
        n_jobs=-1,
        verbose=0,
        random_state=None,
        early_stopping_rounds=5,
        eval_metric="auc",
    ):
        """
        This method initializes the class.

        Args:
            model (sklearn compatible classifier or regressor, sklearn compatible search CV e.g. GridSearchCV, RandomizedSearchCV or BayesSearchCV):
                A model that will be optimized and trained at each round of features elimination. The model must
                support early stopping of training, which is the case for XGBoost and LightGBM, for example. The
                recommended model is [LGBMClassifier](https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMClassifier.html),
                because it by default handles the missing values and categorical variables. This parameter also supports
                any hyperparameter search schema that is consistent with the sklearn API e.g.
                [GridSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.GridSearchCV.html),
                [RandomizedSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.RandomizedSearchCV.html)
                or [BayesSearchCV](https://scikit-optimize.github.io/stable/modules/generated/skopt.BayesSearchCV.html#skopt.BayesSearchCV).
                Note that if a hyperparemeter search model is used, the hyperparameters are tuned without early
                stopping. Early stopping is applied only during the Shapley values estimation for feature
                elimination. We recommend simply passing the model without hyperparameter optimization, or using
                ShapRFECV without early stopping.


            step (int or float, optional):
                Number of lowest importance features removed each round. If it is an int, then each round such number of
                features is discarded. If float, such percentage of remaining features (rounded down) is removed each
                iteration. It is recommended to use float, since it is faster for a large number of features, and slows
                down and becomes more precise towards less features. Note: the last round may remove fewer features in
                order to reach min_features_to_select.
                If columns_to_keep parameter is specified in the fit method, step is the number of features to remove after
                keeping those columns.

            min_features_to_select (int, optional):
                Minimum number of features to be kept. This is a stopping criterion of the feature elimination. By
                default the process stops when one feature is left. If columns_to_keep is specified in the fit method,
                it may override this parameter to the maximum between length of columns_to_keep the two.

            cv (int, cross-validation generator or an iterable, optional):
                Determines the cross-validation splitting strategy. Compatible with sklearn
                [cv parameter](https://scikit-learn.org/stable/modules/generated/sklearn.feature_selection.RFECV.html).
                If None, then cv of 5 is used.

            scoring (string or probatus.utils.Scorer, optional):
                Metric for which the model performance is calculated. It can be either a metric name  aligned with predefined
                [classification scorers names in sklearn](https://scikit-learn.org/stable/modules/model_evaluation.html).
                Another option is using probatus.utils.Scorer to define a custom metric.

            n_jobs (int, optional):
                Number of cores to run in parallel while fitting across folds. None means 1 unless in a
                `joblib.parallel_backend` context. -1 means using all processors.

            verbose (int, optional):
                Controls verbosity of the output:

                - 0 - neither prints nor warnings are shown
                - 1 - only most important warnings
                - 2 - shows all prints and all warnings.

            random_state (int, optional):
                Random state set at each round of feature elimination. If it is None, the results will not be
                reproducible and in random search at each iteration a different hyperparameters might be tested. For
                reproducible results set it to integer.

            early_stopping_rounds (int, optional):
                Number of rounds with constant performance after which the model fitting stops. This is passed to the
                fit method of the model for Shapley values estimation, but not for hyperparameter search. Only
                supported by some models, such as XGBoost and LightGBM.

            eval_metric (str, optional):
                Metric for scoring fitting rounds and activating early stopping. This is passed to the
                fit method of the model for Shapley values estimation, but not for hyperparameter search. Only
                supported by some models, such as [XGBoost](https://xgboost.readthedocs.io/en/latest/parameter.html#learning-task-parameters)
                and [LightGBM](https://lightgbm.readthedocs.io/en/latest/Parameters.html#metric-parameters).
                Note that `eval_metric` is an argument of the model's fit method and it is different from `scoring`.
        """  # noqa
        super().__init__(
            model,
            step=step,
            min_features_to_select=min_features_to_select,
            cv=cv,
            scoring=scoring,
            n_jobs=n_jobs,
            verbose=verbose,
            random_state=random_state,
        )

        if self.search_model and self.verbose > 0:
            warnings.warn(
                "Early stopping will be used only during Shapley value"
                " estimation step, and not for hyperparameter"
                " optimization."
            )

        if not isinstance(early_stopping_rounds, int) or early_stopping_rounds <= 0:
            raise ValueError(
                f"The current value of early_stopping_rounds ="
                f" {early_stopping_rounds} is not allowed."
                f" It needs to be a positive integer."
            )

        self.early_stopping_rounds = early_stopping_rounds
        self.eval_metric = eval_metric

    def _get_fit_params_lightGBM(
        self, X_train, y_train, X_val, y_val, sample_weight=None, train_index=None, val_index=None
    ):
        """Get the fit parameters for for a LightGBM Model.

        Args:

            X_train (pd.DataFrame):
                Train Dataset used in CV.

            y_train (pd.Series):
                Train labels for X.

            X_val (pd.DataFrame):
                Validation Dataset used in CV.

            y_val (pd.Series):
                Validation labels for X.

            sample_weight (pd.Series, np.ndarray, list, optional):
                array-like of shape (n_samples,) - only use if the model you're using supports
                sample weighting (check the corresponding scikit-learn documentation).
                Array of weights that are assigned to individual samples.
                Note that they're only used for fitting of  the model, not during evaluation of metrics.
                If not provided, then each sample is given unit weight.

            train_index (np.array):
                Positions of train folds samples.

            val_index (np.array):
                Positions of validation fold samples.

        Raises:
            ValueError: if the model is not supported.

        Returns:
            dict: fit parameters
        """
        from lightgbm import early_stopping, log_evaluation

        fit_params = {
            "X": X_train,
            "y": y_train,
            "eval_set": [(X_val, y_val)],
            "eval_metric": self.eval_metric,
            "callbacks": [
                early_stopping(self.early_stopping_rounds, first_metric_only=True),
                log_evaluation(1 if self.verbose >= 2 else 0),
            ],
        }

        if sample_weight is not None:
            fit_params["sample_weight"] = sample_weight.iloc[train_index]
            fit_params["eval_sample_weight"] = [sample_weight.iloc[val_index]]

        return fit_params

    def _get_fit_params_XGBoost(
        self, X_train, y_train, X_val, y_val, sample_weight=None, train_index=None, val_index=None
    ):
        """Get the fit parameters for for a XGBoost Model.

        Args:

            X_train (pd.DataFrame):
                Train Dataset used in CV.

            y_train (pd.Series):
                Train labels for X.

            X_val (pd.DataFrame):
                Validation Dataset used in CV.

            y_val (pd.Series):
                Validation labels for X.

            sample_weight (pd.Series, np.ndarray, list, optional):
                array-like of shape (n_samples,) - only use if the model you're using supports
                sample weighting (check the corresponding scikit-learn documentation).
                Array of weights that are assigned to individual samples.
                Note that they're only used for fitting of  the model, not during evaluation of metrics.
                If not provided, then each sample is given unit weight.

            train_index (np.array):
                Positions of train folds samples.

            val_index (np.array):
                Positions of validation fold samples.

        Raises:
            ValueError: if the model is not supported.

        Returns:
            dict: fit parameters
        """
        fit_params = {
            "X": X_train,
            "y": y_train,
            "eval_set": [(X_val, y_val)],
        }
        if sample_weight is not None:
            fit_params["sample_weight"] = sample_weight.iloc[train_index]
            fit_params["eval_sample_weight"] = [sample_weight.iloc[val_index]]

        return fit_params

    def _get_fit_params_CatBoost(
        self, X_train, y_train, X_val, y_val, sample_weight=None, train_index=None, val_index=None
    ):
        """Get the fit parameters for for a CatBoost Model.

        Args:

            X_train (pd.DataFrame):
                Train Dataset used in CV.

            y_train (pd.Series):
                Train labels for X.

            X_val (pd.DataFrame):
                Validation Dataset used in CV.

            y_val (pd.Series):
                Validation labels for X.

            sample_weight (pd.Series, np.ndarray, list, optional):
                array-like of shape (n_samples,) - only use if the model you're using supports
                sample weighting (check the corresponding scikit-learn documentation).
                Array of weights that are assigned to individual samples.
                Note that they're only used for fitting of  the model, not during evaluation of metrics.
                If not provided, then each sample is given unit weight.

            train_index (np.array):
                Positions of train folds samples.

            val_index (np.array):
                Positions of validation fold samples.

        Raises:
            ValueError: if the model is not supported.

        Returns:
            dict: fit parameters
        """
        from catboost import Pool

        cat_features = [col for col in X_train.select_dtypes(include=["category"]).columns]
        fit_params = {
            "X": Pool(X_train, y_train, cat_features=cat_features),
            "eval_set": Pool(X_val, y_val, cat_features=cat_features),
            # Evaluation metric should be passed during initialization
        }
        if sample_weight is not None:
            fit_params["X"].set_weight(sample_weight.iloc[train_index])
            fit_params["eval_set"].set_weight(sample_weight.iloc[val_index])

        return fit_params

    def _get_fit_params(
        self, model, X_train, y_train, X_val, y_val, sample_weight=None, train_index=None, val_index=None
    ):
        """Get the fit parameters for the specified classifier or regressor.

        Args:
            model (classifier or regressor):
                Model to be fitted on the train folds.

            X_train (pd.DataFrame):
                Train Dataset used in CV.

            y_train (pd.Series):
                Train labels for X.

            X_val (pd.DataFrame):
                Validation Dataset used in CV.

            y_val (pd.Series):
                Validation labels for X.

            sample_weight (pd.Series, np.ndarray, list, optional):
                array-like of shape (n_samples,) - only use if the model you're using supports
                sample weighting (check the corresponding scikit-learn documentation).
                Array of weights that are assigned to individual samples.
                Note that they're only used for fitting of  the model, not during evaluation of metrics.
                If not provided, then each sample is given unit weight.

            train_index (np.array):
                Positions of train folds samples.

            val_index (np.array):
                Positions of validation fold samples.

        Raises:
            ValueError: if the model is not supported.

        Returns:
            dict: fit parameters
        """
        # The lightgbm and xgboost imports are temporarily placed here, until the tests on
        # macOS have been fixed.

        try:
            from lightgbm import LGBMModel

            if isinstance(model, LGBMModel):
                return self._get_fit_params_lightGBM(
                    X_train=X_train,
                    y_train=y_train,
                    X_val=X_val,
                    y_val=y_val,
                    sample_weight=sample_weight,
                    train_index=train_index,
                    val_index=val_index,
                )
        except ImportError:
            pass

        try:
            from xgboost.sklearn import XGBModel

            if isinstance(model, XGBModel):
                return self._get_fit_params_XGBoost(
                    X_train=X_train,
                    y_train=y_train,
                    X_val=X_val,
                    y_val=y_val,
                    sample_weight=sample_weight,
                    train_index=train_index,
                    val_index=val_index,
                )
        except ImportError:
            pass

        try:
            from catboost import CatBoost

            if isinstance(model, CatBoost):
                return self._get_fit_params_CatBoost(
                    X_train=X_train,
                    y_train=y_train,
                    X_val=X_val,
                    y_val=y_val,
                    sample_weight=sample_weight,
                    train_index=train_index,
                    val_index=val_index,
                )
        except ImportError:
            pass

        raise ValueError("Model type not supported")

    def _get_feature_shap_values_per_fold(
        self,
        X,
        y,
        model,
        train_index,
        val_index,
        sample_weight=None,
        **shap_kwargs,
    ):
        """
        This function calculates the shap values on validation set, and Train and Val score.

        Args:
            X (pd.DataFrame):
                Dataset used in CV.

            y (pd.Series):
                Labels for X.

            sample_weight (pd.Series, np.ndarray, list, optional):
                array-like of shape (n_samples,) - only use if the model you're using supports
                sample weighting (check the corresponding scikit-learn documentation).
                Array of weights that are assigned to individual samples.
                Note that they're only used for fitting of  the model, not during evaluation of metrics.
                If not provided, then each sample is given unit weight.

            model:
                Classifier or regressor to be fitted on the train folds.

            train_index (np.array):
                Positions of train folds samples.

            val_index (np.array):
                Positions of validation fold samples.

            **shap_kwargs:
                keyword arguments passed to
                [shap.Explainer](https://shap.readthedocs.io/en/latest/generated/shap.Explainer.html#shap.Explainer).
                It also enables `approximate` and `check_additivity` parameters, passed while calculating SHAP values.
                The `approximate=True` causes less accurate, but faster SHAP values calculation, while
                `check_additivity=False` disables the additivity check inside SHAP.
        Returns:
            (np.array, float, float):
                Tuple with the results: Shap Values on validation fold, train score, validation score.
        """
        X_train, X_val = X.iloc[train_index, :], X.iloc[val_index, :]
        y_train, y_val = y.iloc[train_index], y.iloc[val_index]

        fit_params = self._get_fit_params(
            model=model,
            X_train=X_train,
            y_train=y_train,
            X_val=X_val,
            y_val=y_val,
            sample_weight=sample_weight,
            train_index=train_index,
            val_index=val_index,
        )

        # Due to deprecation issues (compatibility with Sklearn) set some params
        # like below, instead of through fit().
        try:
            from xgboost.sklearn import XGBModel

            if isinstance(model, XGBModel):
                model.set_params(eval_metric=self.eval_metric, early_stopping_rounds=self.early_stopping_rounds)
        except ImportError:
            pass

        try:
            from catboost import CatBoost

            if isinstance(model, CatBoost):
                model.set_params(early_stopping_rounds=self.early_stopping_rounds)
        except ImportError:
            pass

        # Train the model
        model = model.fit(**fit_params)

        # Score the model
        score_train = self.scorer.score(model, X_train, y_train)
        score_val = self.scorer.score(model, X_val, y_val)

        # Compute SHAP values
        shap_values = shap_calc(model, X_val, verbose=self.verbose, random_state=self.random_state, **shap_kwargs)
        return shap_values, score_train, score_val



from .resemblance_model import (
    BaseResemblanceModel,
    PermutationImportanceResemblance,
    SHAPImportanceResemblance,
)

__all__ = [
    "BaseResemblanceModel",
    "PermutationImportanceResemblance",
    "SHAPImportanceResemblance",
]



# This approach is adapted from, and explained in: https://calmcode.io/docs/epic.html

import os
from typing import List

import matplotlib
import matplotlib.pyplot as plt
import pytest

import probatus.feature_elimination
import probatus.interpret
import probatus.sample_similarity
import probatus.utils

# Turn off interactive mode in plots
plt.ioff()
matplotlib.use("Agg")

CLASSES_TO_TEST = [
    probatus.feature_elimination.ShapRFECV,
    probatus.interpret.DependencePlotter,
    probatus.sample_similarity.SHAPImportanceResemblance,
    probatus.sample_similarity.PermutationImportanceResemblance,
    probatus.utils.Scorer,
]

CLASSES_TO_TEST_LGBM = [
    probatus.feature_elimination.EarlyStoppingShapRFECV,
]

FUNCTIONS_TO_TEST: List = []


def handle_docstring(doc, indent):
    """
    Check python code in docstring.

    This function will read through the docstring and grab
    the first python code block. It will try to execute it.
    If it fails, the calling test should raise a flag.
    """
    if not doc:
        return
    start = doc.find("```python\n")
    end = doc.find("```\n")
    if start != -1:
        if end != -1:
            code_part = doc[(start + 10) : end].replace(" " * indent, "")
            exec(code_part)


@pytest.mark.parametrize("c", CLASSES_TO_TEST)
def test_class_docstrings(c):
    """
    Take the docstring of a given class.

    The test passes if the usage examples causes no errors.
    """
    handle_docstring(c.__doc__, indent=4)


@pytest.mark.skipif(os.environ.get("SKIP_LIGHTGBM") == "true", reason="LightGBM tests disabled")
@pytest.mark.parametrize("c", CLASSES_TO_TEST_LGBM)
def test_class_docstrings_lgbm(c):
    """
    Take the docstring of a given class which uses LightGBM.

    The test passes if the usage examples causes no errors.

    The test is skipped if the environment does not support LightGBM correctly, such as macos.
    """
    handle_docstring(c.__doc__, indent=4)


@pytest.mark.parametrize("f", FUNCTIONS_TO_TEST)
def test_function_docstrings(f):
    """
    Take the docstring of every function.

    The test passes if the usage examples causes no errors.
    """
    handle_docstring(f.__doc__, indent=4)



import os
from pathlib import Path

import pytest
import nbformat
from nbconvert.preprocessors import ExecutePreprocessor

TIMEOUT_SECONDS = 1800
PATH_NOTEBOOKS = [str(path) for path in Path("docs").glob("*/*.ipynb")]
NB_FLAG = os.environ.get("TEST_NOTEBOOKS")  # Turn on tests by setting TEST_NOTEBOOKS = 1
TEST_NOTEBOOKS = False if NB_FLAG == "1" else True


@pytest.mark.parametrize("notebook_path", PATH_NOTEBOOKS)
@pytest.mark.skipif(TEST_NOTEBOOKS, reason="Skip notebook tests if TEST_NOTEBOOK isn't set")
def test_notebook(notebook_path: str) -> None:
    """Run a notebook and check no exception is raised."""
    with open(notebook_path) as f:
        nb = nbformat.read(f, as_version=4)

    ep = ExecutePreprocessor(timeout=TIMEOUT_SECONDS, kernel_name="python3")
    ep.preprocess(nb, {"metadata": {"path": str(Path(notebook_path).parent)}})



from .exceptions import NotFittedError
from .arrayfuncs import (
    assure_pandas_df,
    assure_pandas_series,
    preprocess_data,
    preprocess_labels,
)
from .scoring import Scorer, get_single_scorer
from .shap_helpers import shap_calc, shap_to_df, calculate_shap_importance
from ._utils import assure_list_of_strings
from .base_class_interface import BaseFitComputeClass, BaseFitComputePlotClass

__all__ = [
    "assure_list_of_strings",
    "assure_pandas_df",
    "assure_pandas_series",
    "preprocess_data",
    "preprocess_labels",
    "BaseFitComputeClass",
    "BaseFitComputePlotClass",
    "NotFittedError",
    "get_single_scorer",
    "Scorer",
    "shap_calc",
    "shap_to_df",
    "calculate_shap_importance",
]



import warnings

import numpy as np
import pandas as pd
from shap import Explainer
from shap.explainers import TreeExplainer
from shap.utils import sample
from sklearn.pipeline import Pipeline


def shap_calc(
    model,
    X,
    return_explainer=False,
    verbose=0,
    random_state=None,
    sample_size=100,
    approximate=False,
    check_additivity=True,
    **shap_kwargs,
):
    """
    Helper function to calculate the shapley values for a given model.

    Args:
        model (model):
            Trained model.

        X (pd.DataFrame or np.ndarray):
            features set.

        return_explainer (boolean):
            if True, returns a a tuple (shap_values, explainer).

        verbose (int, optional):
            Controls verbosity of the output:

            - 0 - neither prints nor warnings are shown
            - 1 - only most important warnings
            - 2 - shows all prints and all warnings.

        random_state (int, optional):
            Random state set for the nr of samples. If it is None, the results will not be reproducible. For
            reproducible results set it to an integer.

        approximate (boolean):
            if True uses shap approximations - less accurate, but very fast. It applies to tree-based explainers only.

        check_additivity (boolean):
            if False SHAP will disable the additivity check for tree-based models.

        **shap_kwargs: kwargs of the shap.Explainer

    Returns:
        (np.ndarray or tuple(np.ndarray, shap.Explainer)):
            shapley_values for the model, optionally also returns the explainer.

    """
    if isinstance(model, Pipeline):
        raise TypeError(
            "The provided model is a Pipeline. Unfortunately, the features based on SHAP do not support "
            "pipelines, because they cannot be used in combination with shap.Explainer. Please apply any "
            "data transformations before running the probatus module."
        )

    # Suppress warnings regarding XGboost and Lightgbm models.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore" if verbose <= 1 else "default")

        # For tree explainers, do not pass masker when feature_perturbation is
        # tree_path_dependent, or when X contains categorical features
        # related to issue:
        # https://github.com/slundberg/shap/issues/480
        if shap_kwargs.get("feature_perturbation") == "tree_path_dependent" or X.select_dtypes("category").shape[1] > 0:
            # Calculate Shap values.
            explainer = Explainer(model, seed=random_state, **shap_kwargs)
        else:
            # Create the background data,required for non tree based models.
            # A single datapoint can passed as mask
            # (https://github.com/slundberg/shap/issues/955#issuecomment-569837201)
            if X.shape[0] < sample_size:
                sample_size = int(np.ceil(X.shape[0] * 0.2))
            else:
                pass
            mask = sample(X, sample_size, random_state=random_state)
            explainer = Explainer(model, seed=random_state, masker=mask, **shap_kwargs)

        # For tree-explainers allow for using check_additivity and approximate arguments
        if isinstance(explainer, TreeExplainer):
            shap_values = explainer.shap_values(X, check_additivity=check_additivity, approximate=approximate)

            # From SHAP version 0.43+ https://github.com/shap/shap/pull/3121 required to
            # get the second dimension of calculated Shap values.
            if not isinstance(shap_values, list) and len(shap_values.shape) == 3:
                shap_values = shap_values[:, :, 1]
        else:
            # Calculate Shap values
            shap_values = explainer.shap_values(X)

        if isinstance(shap_values, list) and len(shap_values) == 2:
            warnings.warn(
                "Shap values are related to the output probabilities of class 1 for this model, instead of " "log odds."
            )
            shap_values = shap_values[1]

    if return_explainer:
        return shap_values, explainer
    return shap_values


def shap_to_df(model, X, precalc_shap=None, **kwargs):
    """
    Calculates the shap values and return the pandas DataFrame with the columns and the index of the original.

    Args:
        model (model):
            Pretrained model (Random Forest of XGBoost at the moment).

        X (pd.DataFrame or np.ndarray):
            Dataset on which the SHAP importance is calculated.

        precalc_shap (np.array):
            Precalculated SHAP values. If None, they are computed.

        **kwargs: for the function shap_calc

    Returns:
        (pd.DataFrame):
            Dataframe with SHAP feature importance per features on X dataset.
    """
    shap_values = precalc_shap if precalc_shap is not None else shap_calc(model, X, **kwargs)

    try:
        return pd.DataFrame(shap_values, columns=X.columns, index=X.index)
    except AttributeError:
        if isinstance(X, np.ndarray) and len(X.shape) == 2:
            return pd.DataFrame(shap_values, columns=[f"col_{ix}" for ix in range(X.shape[1])])
        else:
            raise TypeError("X must be a dataframe or a 2d array")


def calculate_shap_importance(shap_values, columns, output_columns_suffix="", shap_variance_penalty_factor=None):
    """
    Returns the average shapley value for each column of the dataframe, as well as the average absolute shap value.

    Args:
        shap_values (np.array):
            Shap values.

        columns (list of str):
            Feature names.

        output_columns_suffix (str, optional):
            Suffix to be added at the end of column names in the output.

        shap_variance_penalty_factor (int or float, optional):
            Apply aggregation penalty when computing average of shap values for a given feature.
            Results in a preference for features that have smaller standard deviation of shap
            values (more coherent shap importance). Recommend value 0.5 - 1.0.
            Formula: penalized_shap_mean = (mean_shap - (std_shap * shap_variance_penalty_factor))

    Returns:
        (pd.DataFrame):
            Mean absolute shap values and Mean shap values of features.

    """
    if shap_variance_penalty_factor is None or shap_variance_penalty_factor < 0:
        shap_variance_penalty_factor = 0
    elif not isinstance(shap_variance_penalty_factor, (float, int)):
        warnings.warn(
            "shap_variance_penalty_factor must be None, int, or float. Setting shap_variance_penalty_factor = 0"
        )
        shap_variance_penalty_factor = 0

    abs_shap_values = np.abs(shap_values)
    if np.ndim(shap_values) > 2:  # multi-class case
        sum_abs_shap = np.sum(abs_shap_values, axis=0)
        sum_shap = np.sum(shap_values, axis=0)
        shap_abs_mean = np.mean(sum_abs_shap, axis=0)
        shap_mean = np.mean(sum_shap, axis=0)
        penalized_shap_abs_mean = shap_abs_mean - (np.std(sum_abs_shap, axis=0) * shap_variance_penalty_factor)
    else:
        # Find average shap importance for neg and pos class
        shap_abs_mean = np.mean(abs_shap_values, axis=0)
        shap_mean = np.mean(shap_values, axis=0)
        penalized_shap_abs_mean = shap_abs_mean - (np.std(abs_shap_values, axis=0) * shap_variance_penalty_factor)

    # Prepare the values in a df and set the correct column types
    importance_df = pd.DataFrame(
        {
            f"mean_abs_shap_value{output_columns_suffix}": shap_abs_mean,
            f"mean_shap_value{output_columns_suffix}": shap_mean,
            f"penalized_mean_abs_shap_value{output_columns_suffix}": penalized_shap_abs_mean,
        },
        index=columns,
    ).astype(float)

    importance_df = importance_df.sort_values(f"penalized_mean_abs_shap_value{output_columns_suffix}", ascending=False)

    # Drop penalized column
    importance_df = importance_df.drop(columns=[f"penalized_mean_abs_shap_value{output_columns_suffix}"])

    return importance_df



import numpy as np
import pandas as pd
import pytest

from probatus.utils import assure_pandas_df, preprocess_data, preprocess_labels


@pytest.fixture(scope="function")
def expected_df_2d():
    return pd.DataFrame({0: [1, 2], 1: [2, 3], 2: [3, 4]})


@pytest.fixture(scope="function")
def expected_df():
    return pd.DataFrame({0: [1, 2, 3]})


def test_assure_pandas_df_list(expected_df):
    x = [1, 2, 3]
    x_df = assure_pandas_df(x)
    pd.testing.assert_frame_equal(x_df, expected_df)


def test_assure_pandas_df_list_of_lists(expected_df_2d):
    x = [[1, 2, 3], [2, 3, 4]]
    x_df = assure_pandas_df(x)
    pd.testing.assert_frame_equal(x_df, expected_df_2d)


def test_assure_pandas_df_series(expected_df):
    x = pd.Series([1, 2, 3])
    x_df = assure_pandas_df(x)
    pd.testing.assert_frame_equal(x_df, expected_df)


def test_assure_pandas_df_array(expected_df, expected_df_2d):
    x = np.array([[1, 2, 3], [2, 3, 4]], dtype="int64")
    x_df = assure_pandas_df(x)
    pd.testing.assert_frame_equal(x_df, expected_df_2d)

    x = np.array([1, 2, 3], dtype="int64")
    x_df = assure_pandas_df(x)
    pd.testing.assert_frame_equal(x_df, expected_df)


def test_assure_pandas_df_df(expected_df_2d):
    x = pd.DataFrame([[1, 2, 3], [2, 3, 4]])
    x_df = assure_pandas_df(x)
    pd.testing.assert_frame_equal(x_df, expected_df_2d)


def test_assure_pandas_df_types():
    with pytest.raises(TypeError):
        assure_pandas_df("Test")
    with pytest.raises(TypeError):
        assure_pandas_df(5)


def test_preprocess_labels():
    y1 = pd.Series([1, 0, 1, 0, 1])
    index_1 = np.array([5, 4, 3, 2, 1])

    y1_output = preprocess_labels(y1, y_name="y1", index=index_1, verbose=2)
    pd.testing.assert_series_equal(y1_output, pd.Series([1, 0, 1, 0, 1], index=index_1))

    y2 = [False, False, False, False, False]
    y2_output = preprocess_labels(y2, y_name="y2", verbose=2)
    pd.testing.assert_series_equal(y2_output, pd.Series(y2))

    y3 = np.array([0, 1, 2, 3, 4])
    y3_output = preprocess_labels(y3, y_name="y3", verbose=2)
    pd.testing.assert_series_equal(y3_output, pd.Series(y3))

    y4 = pd.Series(["2", "1", "3", "2", "1"])
    index4 = pd.Index([0, 2, 1, 3, 4])
    y4_output = preprocess_labels(y4, y_name="y4", index=index4, verbose=0)
    pd.testing.assert_series_equal(y4_output, pd.Series(["2", "3", "1", "2", "1"], index=index4))


def test_preprocess_data():
    X1 = pd.DataFrame({"cat": ["a", "b", "c"], "missing": [1, np.nan, 2], "num_1": [1, 2, 3]})

    target_column_names_X1 = ["1", "2", "3"]
    X1_expected_output = pd.DataFrame({"1": ["a", "b", "c"], "2": [1, np.nan, 2], "3": [1, 2, 3]})

    X1_expected_output["1"] = X1_expected_output["1"].astype("category")
    X1_output, output_column_names_X1 = preprocess_data(X1, X_name="X1", column_names=target_column_names_X1, verbose=2)
    assert target_column_names_X1 == output_column_names_X1
    pd.testing.assert_frame_equal(X1_output, X1_expected_output)

    X2 = np.array([[1, 3, 2], [1, 2, 2], [1, 2, 3]])

    target_column_names_X1 = [0, 1, 2]
    X2_expected_output = pd.DataFrame(X2, columns=target_column_names_X1)
    X2_output, output_column_names_X2 = preprocess_data(X2, X_name="X2", column_names=None, verbose=2)

    assert target_column_names_X1 == output_column_names_X2
    pd.testing.assert_frame_equal(X2_output, X2_expected_output)



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from shap import summary_plot
from shap.plots._waterfall import waterfall_legacy

from probatus.interpret import DependencePlotter
from probatus.utils import (
    BaseFitComputePlotClass,
    assure_list_of_strings,
    calculate_shap_importance,
    preprocess_data,
    preprocess_labels,
    get_single_scorer,
    shap_calc,
)


class ShapModelInterpreter(BaseFitComputePlotClass):
    """
    This class is a wrapper that allows to easily analyse a model's features.

    It allows us to plot SHAP feature importance,
        SHAP summary plot and SHAP dependence plots.

    Example:
    ```python
    from sklearn.datasets import make_classification
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.model_selection import train_test_split
    from probatus.interpret import ShapModelInterpreter
    import numpy as np
    import pandas as pd

    feature_names = ['f1', 'f2', 'f3', 'f4']

    # Prepare two samples
    X, y = make_classification(n_samples=5000, n_features=4, random_state=0)
    X = pd.DataFrame(X, columns=feature_names)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)

    # Prepare and fit model. Remember about class_weight="balanced" or an equivalent.
    model = RandomForestClassifier(class_weight='balanced', n_estimators = 100, max_depth=2, random_state=0)
    model.fit(X_train, y_train)

    # Train ShapModelInterpreter
    shap_interpreter = ShapModelInterpreter(model)
    feature_importance = shap_interpreter.fit_compute(X_train, X_test, y_train, y_test)

    # Make plots
    ax1 = shap_interpreter.plot('importance')
    ax2 = shap_interpreter.plot('summary')
    ax3 = shap_interpreter.plot('dependence', target_columns=['f1', 'f2'])
    ax4 = shap_interpreter.plot('sample', samples_index=[X_test.index.tolist()[0]])
    ```

    <img src="../img/model_interpret_importance.png" width="320" />
    <img src="../img/model_interpret_summary.png" width="320" />
    <img src="../img/model_interpret_dep.png" width="320" />
    <img src="../img/model_interpret_sample.png" width="320" />
    """

    def __init__(self, model, scoring="roc_auc", verbose=0, random_state=None):
        """
        Initializes the class.

        Args:
            model (classifier or regressor):
                Model fitted on X_train.

            scoring (string or probatus.utils.Scorer, optional):
                Metric for which the model performance is calculated. It can be either a metric name  aligned with
                predefined classification scorers names in sklearn
                ([link](https://scikit-learn.org/stable/modules/model_evaluation.html)).
                Another option is using probatus.utils.Scorer to define a custom metric.

            verbose (int, optional):
                Controls verbosity of the output:

                - 0 - neither prints nor warnings are shown
                - 1 - only most important warnings
                - 2 - shows all prints and all warnings.

            random_state (int, optional):
                Random state set for the nr of samples. If it is None, the results will not be reproducible. For
                reproducible results set it to an integer.
        """
        self.model = model
        self.scorer = get_single_scorer(scoring)
        self.verbose = verbose
        self.random_state = random_state

    def fit(
        self,
        X_train,
        X_test,
        y_train,
        y_test,
        column_names=None,
        class_names=None,
        **shap_kwargs,
    ):
        """
        Fits the object and calculates the shap values for the provided datasets.

        Args:
            X_train (pd.DataFrame):
                Dataframe containing training data.

            X_test (pd.DataFrame):
                Dataframe containing test data.

            y_train (pd.Series):
                Series of labels for train data.

            y_test (pd.Series):
                Series of labels for test data.

            column_names (None, or list of str, optional):
                List of feature names for the dataset. If None, then column names from the X_train dataframe are used.

            class_names (None, or list of str, optional):
                List of class names e.g. ['neg', 'pos']. If none, the default ['Negative Class', 'Positive Class'] are
                used.

            **shap_kwargs:
                keyword arguments passed to
                [shap.Explainer](https://shap.readthedocs.io/en/latest/generated/shap.Explainer.html#shap.Explainer).
                It also enables `approximate` and `check_additivity` parameters, passed while calculating SHAP values.
                The `approximate=True` causes less accurate, but faster SHAP values calculation, while
                `check_additivity=False` disables the additivity check inside SHAP.
        """

        self.X_train, self.column_names = preprocess_data(
            X_train, X_name="X_train", column_names=column_names, verbose=self.verbose
        )
        self.X_test, _ = preprocess_data(X_test, X_name="X_test", column_names=column_names, verbose=self.verbose)
        self.y_train = preprocess_labels(y_train, y_name="y_train", index=self.X_train.index, verbose=self.verbose)
        self.y_test = preprocess_labels(y_test, y_name="y_test", index=self.X_test.index, verbose=self.verbose)

        # Set class names
        self.class_names = class_names
        if self.class_names is None:
            self.class_names = ["Negative Class", "Positive Class"]

        # Calculate Metrics
        self.train_score = self.scorer.score(self.model, self.X_train, self.y_train)
        self.test_score = self.scorer.score(self.model, self.X_test, self.y_test)

        self.results_text = (
            f"Train {self.scorer.metric_name}: {np.round(self.train_score, 3)},\n"
            f"Test {self.scorer.metric_name}: {np.round(self.test_score, 3)}."
        )

        (
            self.shap_values_train,
            self.expected_value_train,
            self.tdp_train,
        ) = self._prep_shap_related_variables(
            model=self.model,
            X=self.X_train,
            y=self.y_train,
            column_names=self.column_names,
            class_names=self.class_names,
            verbose=self.verbose,
            random_state=self.random_state,
            **shap_kwargs,
        )

        (
            self.shap_values_test,
            self.expected_value_test,
            self.tdp_test,
        ) = self._prep_shap_related_variables(
            model=self.model,
            X=self.X_test,
            y=self.y_test,
            column_names=self.column_names,
            class_names=self.class_names,
            verbose=self.verbose,
            random_state=self.random_state,
            **shap_kwargs,
        )

        self.fitted = True

    @staticmethod
    def _prep_shap_related_variables(
        model,
        X,
        y,
        approximate=False,
        verbose=0,
        random_state=None,
        column_names=None,
        class_names=None,
        **shap_kwargs,
    ):
        """
        The function prepares the variables related to shap that are used to interpret the model.

        Returns:
            (np.array, int, DependencePlotter):
                Shap values, expected value of the explainer, and fitted TreeDependencePlotter for a given dataset.
        """
        shap_values, explainer = shap_calc(
            model,
            X,
            approximate=approximate,
            verbose=verbose,
            random_state=random_state,
            return_explainer=True,
            **shap_kwargs,
        )

        expected_value = explainer.expected_value

        # For sklearn models the expected values consists of two elements (negative_class and positive_class)
        if isinstance(expected_value, list) or isinstance(expected_value, np.ndarray):
            expected_value = expected_value[1]

        # Initialize tree dependence plotter
        tdp = DependencePlotter(model, verbose=verbose).fit(
            X,
            y,
            column_names=column_names,
            class_names=class_names,
            precalc_shap=shap_values,
        )
        return shap_values, expected_value, tdp

    def compute(self, return_scores=False, shap_variance_penalty_factor=None):
        """
        Computes the DataFrame that presents the importance of each feature.

        Args:
            return_scores (bool, optional):
                Flag indicating whether the method should return the train and test score of the model, together with
                the model interpretation report. If true, the output of this method is a tuple of DataFrame, float,
                float.

            shap_variance_penalty_factor (int or float, optional):
                Apply aggregation penalty when computing average of shap values for a given feature.
                Results in a preference for features that have smaller standard deviation of shap
                values (more coherent shap importance). Recommend value 0.5 - 1.0.
                Formula: penalized_shap_mean = (mean_shap - (std_shap * shap_variance_penalty_factor))

        Returns:
            (pd.DataFrame or tuple(pd.DataFrame, float, float)):
                Dataframe with SHAP feature importance, or tuple containing the dataframe, train and test scores of the
                model.
        """
        self._check_if_fitted()

        # Compute SHAP importance
        self.importance_df_train = calculate_shap_importance(
            self.shap_values_train,
            self.column_names,
            output_columns_suffix="_train",
            shap_variance_penalty_factor=shap_variance_penalty_factor,
        )

        self.importance_df_test = calculate_shap_importance(
            self.shap_values_test,
            self.column_names,
            output_columns_suffix="_test",
            shap_variance_penalty_factor=shap_variance_penalty_factor,
        )

        # Concatenate the train and test, sort by test set importance and reorder the columns
        self.importance_df = pd.concat([self.importance_df_train, self.importance_df_test], axis=1).sort_values(
            "mean_abs_shap_value_test", ascending=False
        )[
            [
                "mean_abs_shap_value_test",
                "mean_abs_shap_value_train",
                "mean_shap_value_test",
                "mean_shap_value_train",
            ]
        ]

        if return_scores:
            return self.importance_df, self.train_score, self.test_score
        else:
            return self.importance_df

    def fit_compute(
        self,
        X_train,
        X_test,
        y_train,
        y_test,
        column_names=None,
        class_names=None,
        return_scores=False,
        shap_variance_penalty_factor=None,
        **shap_kwargs,
    ):
        """
        Fits the object and calculates the shap values for the provided datasets.

        Args:
            X_train (pd.DataFrame):
                Dataframe containing training data.

            X_test (pd.DataFrame):
                Dataframe containing test data.

            y_train (pd.Series):
                Series of labels for train data.

            y_test (pd.Series):
                Series of labels for test data.

            column_names (None, or list of str, optional):
                List of feature names for the dataset.
                If None, then column names from the X_train dataframe are used.

            class_names (None, or list of str, optional):
                List of class names e.g. ['neg', 'pos'].
                If none, the default ['Negative Class', 'Positive Class'] are
                used.

            return_scores (bool, optional):
                Flag indicating whether the method should return
                the train and test score of the model,
                together with the model interpretation report. If true,
                the output of this method is a tuple of DataFrame, float,
                float.

            shap_variance_penalty_factor (int or float, optional):
                Apply aggregation penalty when computing average of shap values for a given feature.
                Results in a preference for features that have smaller standard deviation of shap
                values (more coherent shap importance). Recommend value 0.5 - 1.0.
                Formula: penalized_shap_mean = (mean_shap - (std_shap * shap_variance_penalty_factor))

            **shap_kwargs:
                keyword arguments passed to
                [shap.Explainer](https://shap.readthedocs.io/en/latest/generated/shap.Explainer.html#shap.Explainer).
                It also enables `approximate` and `check_additivity` parameters, passed while calculating SHAP values.
                The `approximate=True` causes less accurate, but faster SHAP values calculation, while
                `check_additivity=False` disables the additivity check inside SHAP.

        Returns:
            (pd.DataFrame or tuple(pd.DataFrame, float, float)):
                Dataframe with SHAP feature importance, or tuple containing the dataframe, train and test scores of the
                model.
        """
        self.fit(
            X_train=X_train,
            X_test=X_test,
            y_train=y_train,
            y_test=y_test,
            column_names=column_names,
            class_names=class_names,
            **shap_kwargs,
        )
        return self.compute(return_scores=return_scores, shap_variance_penalty_factor=shap_variance_penalty_factor)

    def plot(self, plot_type, target_set="test", target_columns=None, samples_index=None, show=True, **plot_kwargs):
        """
        Plots the appropriate SHAP plot.

        Args:
            plot_type (str):
                One of the following:

                - `'importance'`: Feature importance plot, SHAP bar summary plot
                - `'summary'`: SHAP Summary plot
                - `'dependence'`: Dependence plot for each feature
                - `'sample'`: Explanation of a given sample in the test data

            target_set (str, optional):
                The set for which the plot should be generated, either `train` or `test` set. We recommend using test
                set, because it is not biased by model training. The train set plots are mainly used to compare with the
                test set plots, whether there are significant differences, which indicate shift in data distribution.

            target_columns (None, str or list of str, optional):
                List of features names, for which the plots should be generated. If None, all features will be plotted.

            samples_index (None, int, list or pd.Index, optional):
                Index of samples to be explained if the `plot_type=sample`.

            show (bool, optional):
                If True, the plots are showed to the user, otherwise they are not shown. Not showing plot can be useful,
                when you want to edit the returned axis, before showing it.

            **plot_kwargs:
                Keyword arguments passed to the plot method. For 'importance' and 'summary' plot_type, the kwargs are
                passed to shap.summary_plot, for 'dependence' plot_type, they are passed to
                probatus.interpret.DependencePlotter.plot method.

        Returns:
            (matplotlib.axes or list(matplotlib.axes)):
                An Axes with the plot, or list of axes when multiple plots are returned.
        """
        # Choose correct columns
        if target_columns is None:
            target_columns = self.column_names

        target_columns = assure_list_of_strings(target_columns, "target_columns")
        target_columns_indices = [self.column_names.index(target_column) for target_column in target_columns]

        # Choose the correct dataset
        if target_set == "test":
            target_X = self.X_test
            target_shap_values = self.shap_values_test
            target_tdp = self.tdp_test
            target_expected_value = self.expected_value_test
        elif target_set == "train":
            target_X = self.X_train
            target_shap_values = self.shap_values_train
            target_tdp = self.tdp_train
            target_expected_value = self.expected_value_train
        else:
            raise (ValueError('The target_set parameter can be either "train" or "test".'))

        if plot_type in ["importance", "summary"]:
            target_X = target_X[target_columns]
            target_shap_values = target_shap_values[:, target_columns_indices]
            # Set summary plot settings
            if plot_type == "importance":
                plot_type = "bar"
                plot_title = f"SHAP Feature Importance for {target_set} set"
            else:
                plot_type = "dot"
                plot_title = f"SHAP Summary plot for {target_set} set"

            summary_plot(
                target_shap_values,
                target_X,
                plot_type=plot_type,
                class_names=self.class_names,
                show=False,
                **plot_kwargs,
            )
            ax = plt.gca()
            ax.set_title(plot_title)

            ax.annotate(
                self.results_text,
                (0, 0),
                (0, -50),
                fontsize=12,
                xycoords="axes fraction",
                textcoords="offset points",
                va="top",
            )
            if show:
                plt.show()
            else:
                plt.close()
        elif plot_type == "dependence":
            ax = []
            for feature_name in target_columns:
                ax.append(target_tdp.plot(feature=feature_name, figsize=(10, 7), show=show, **plot_kwargs))

        elif plot_type == "sample":
            # Ensure the correct samples_index type
            if samples_index is None:
                raise (ValueError("For sample plot, you need to specify the samples_index be plotted plot"))
            elif isinstance(samples_index, int) or isinstance(samples_index, str):
                samples_index = [samples_index]
            elif not (isinstance(samples_index, list) or isinstance(samples_index, pd.Index)):
                raise (TypeError("sample_index must be one of the following: int, str, list or pd.Index"))

            ax = []
            for sample_index in samples_index:
                sample_loc = target_X.index.get_loc(sample_index)

                waterfall_legacy(
                    target_expected_value,
                    target_shap_values[sample_loc, :],
                    target_X.loc[sample_index],
                    show=False,
                    **plot_kwargs,
                )

                plot_title = f"SHAP Sample Explanation of {target_set} sample for index={sample_index}"
                current_ax = plt.gca()
                current_ax.set_title(plot_title)
                ax.append(current_ax)
                if show:
                    plt.show()
                else:
                    plt.close()
        else:
            raise ValueError("Wrong plot type, select from 'importance', 'summary', or 'dependence'")

        if isinstance(ax, list) and len(ax) == 1:
            ax = ax[0]
        return ax



from abc import ABC, abstractmethod

from probatus.utils import NotFittedError


class BaseFitComputeClass(ABC):
    """
    Placeholder that must be overwritten by subclass.
    """

    fitted = False

    def _check_if_fitted(self):
        """
        Checks if object has been fitted. If not, NotFittedError is raised.
        """
        if not self.fitted:
            raise (NotFittedError("The object has not been fitted. Please run fit() method first"))

    @abstractmethod
    def fit(self, *args, **kwargs):
        """
        Placeholder that must be overwritten by subclass.
        """
        pass

    @abstractmethod
    def compute(self, *args, **kwargs):
        """
        Placeholder that must be overwritten by subclass.
        """
        pass

    @abstractmethod
    def fit_compute(self, *args, **kwargs):
        """
        Placeholder that must be overwritten by subclass.
        """
        pass


class BaseFitComputePlotClass(BaseFitComputeClass):
    """
    Base class.
    """

    @abstractmethod
    def plot(self, *args, **kwargs):
        """
        Placeholder method for plotting.
        """
        pass



import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytest
from sklearn.ensemble import RandomForestClassifier

from probatus.interpret.shap_dependence import DependencePlotter
from probatus.utils.exceptions import NotFittedError

# Turn off interactive mode in plots
plt.ioff()
matplotlib.use("Agg")


@pytest.fixture(scope="function")
def X_y():
    return (
        pd.DataFrame(
            [
                [1.72568193, 2.21070436, 1.46039061],
                [-1.48382902, 2.88364928, 0.22323996],
                [-0.44947744, 0.85434638, -2.54486421],
                [-1.38101231, 1.77505901, -1.36000132],
                [-0.18261804, -0.25829609, 1.46925993],
                [0.27514902, 0.09608222, 0.7221381],
                [-0.27264455, 1.99366793, -2.62161046],
                [-2.81587587, 3.46459717, -0.11740999],
                [1.48374489, 0.79662903, 1.18898706],
                [-1.27251335, -1.57344342, -0.39540133],
                [0.31532891, 0.38299269, 1.29998754],
                [-2.10917352, -0.70033132, -0.89922129],
                [-2.14396343, -0.44549774, -1.80572922],
                [-3.4503348, 3.43476247, -0.74957725],
                [-1.25945582, -1.7234203, -0.77435353],
            ]
        ),
        pd.Series([1, 1, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0, 0, 0, 0]),
    )


@pytest.fixture(scope="function")
def expected_shap_vals():
    return pd.DataFrame(
        [
            [0.176667, 0.005833, 0.284167],
            [-0.042020, 0.224520, 0.284167],
            [-0.092020, -0.135480, -0.205833],
            [-0.092020, -0.135480, -0.205833],
            [0.002424, 0.000909, 0.263333],
            [0.176667, 0.105833, 0.284167],
            [-0.092020, -0.135480, -0.205833],
            [-0.028687, 0.311187, 0.184167],
            [0.176667, 0.005833, 0.284167],
            [-0.092020, -0.164646, -0.076667],
            [0.176667, 0.105833, 0.284167],
            [-0.092020, -0.164646, -0.176667],
            [-0.092020, -0.164646, -0.176667],
            [-0.108687, 0.081187, -0.205833],
            [-0.092020, -0.164646, -0.176667],
        ]
    )


@pytest.fixture(scope="function")
def model(X_y, random_state):
    X, y = X_y

    model = RandomForestClassifier(random_state=random_state, n_estimators=10, max_depth=5)

    model.fit(X, y)
    return model


@pytest.fixture(scope="function")
def expected_feat_importances():
    return pd.DataFrame(
        {
            "Feature Name": {0: 2, 1: 1, 2: 0},
            "Shap absolute importance": {0: 0.2199, 1: 0.1271, 2: 0.1022},
            "Shap signed importance": {0: 0.0292, 1: -0.0149, 2: -0.0076},
        }
    )


def test_not_fitted(model, random_state):
    plotter = DependencePlotter(model, random_state)
    assert plotter.fitted is False


def test_fit_complex(complex_data_split, complex_fitted_lightgbm, random_state):
    _, X_test, _, y_test = complex_data_split

    plotter = DependencePlotter(complex_fitted_lightgbm, random_state=random_state)

    plotter.fit(X_test, y_test)

    pd.testing.assert_frame_equal(plotter.X, X_test)
    pd.testing.assert_series_equal(plotter.y, pd.Series(y_test, index=X_test.index))
    assert plotter.fitted is True

    # Check if plotting does not cause errors
    _ = plotter.plot(feature="f2_missing", show=False)


def test_get_X_y_shap_with_q_cut_normal(X_y, model, random_state):
    X, y = X_y

    plotter = DependencePlotter(model, random_state).fit(X, y)
    plotter.min_q, plotter.max_q = 0, 1

    X_cut, y_cut, _ = plotter._get_X_y_shap_with_q_cut(0)
    assert np.isclose(X[0], X_cut).all()
    assert y.equals(y_cut)

    plotter.min_q = 0.2
    plotter.max_q = 0.8

    X_cut, y_cut, _ = plotter._get_X_y_shap_with_q_cut(0)
    assert np.isclose(
        X_cut,
        [
            -1.48382902,
            -0.44947744,
            -1.38101231,
            -0.18261804,
            0.27514902,
            -0.27264455,
            -1.27251335,
            -2.10917352,
            -1.25945582,
        ],
    ).all()
    assert np.equal(y_cut.values, [1, 0, 0, 1, 1, 0, 0, 0, 0]).all()


def test_get_X_y_shap_with_q_cut_unfitted(model, random_state):
    plotter = DependencePlotter(model, random_state)
    with pytest.raises(NotFittedError):
        plotter._get_X_y_shap_with_q_cut(0)


def test_get_X_y_shap_with_q_cut_input(X_y, model, random_state):
    plotter = DependencePlotter(model, random_state).fit(X_y[0], X_y[1])
    with pytest.raises(ValueError):
        plotter._get_X_y_shap_with_q_cut("not a feature")


def test_plot_normal(X_y, model, random_state):
    plotter = DependencePlotter(model, random_state).fit(X_y[0], X_y[1])
    _ = plotter.plot(feature=0)


def test_plot_class_names(X_y, model, random_state):
    plotter = DependencePlotter(model, random_state).fit(X_y[0], X_y[1], class_names=["a", "b"])
    _ = plotter.plot(feature=0)
    assert plotter.class_names == ["a", "b"]


def test_plot_input(X_y, model, random_state):
    plotter = DependencePlotter(model, random_state).fit(X_y[0], X_y[1])
    with pytest.raises(ValueError):
        plotter.plot(feature="not a feature")
    with pytest.raises(TypeError):
        plotter.plot(feature=0, bins=5.0)
    with pytest.raises(ValueError):
        plotter.plot(feature=0, min_q=1, max_q=0)


def test__repr__(model, random_state):
    """
    Test string representation.
    """
    plotter = DependencePlotter(model, random_state)
    assert str(plotter) == "Shap dependence plotter for RandomForestClassifier"



import pandas as pd
import pytest
from lightgbm import LGBMClassifier
from sklearn.datasets import load_diabetes, make_classification
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import RandomizedSearchCV, StratifiedGroupKFold, StratifiedKFold
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.svm import SVC
from xgboost import XGBClassifier, XGBRegressor

from probatus.feature_elimination import EarlyStoppingShapRFECV, ShapRFECV
from probatus.utils import preprocess_labels


@pytest.fixture(scope="function")
def X():
    return pd.DataFrame(
        {
            "col_1": [1, 1, 1, 1, 1, 1, 1, 0],
            "col_2": [0, 0, 0, 0, 0, 0, 0, 1],
            "col_3": [1, 0, 1, 0, 1, 0, 1, 0],
        },
        index=[1, 2, 3, 4, 5, 6, 7, 8],
    )


@pytest.fixture(scope="function")
def y():
    return pd.Series([1, 0, 1, 0, 1, 0, 1, 0], index=[1, 2, 3, 4, 5, 6, 7, 8])


@pytest.fixture(scope="function")
def sample_weight():
    return pd.Series([1, 1, 1, 1, 1, 1, 1, 1], index=[1, 2, 3, 4, 5, 6, 7, 8])


@pytest.fixture(scope="function")
def groups():
    return pd.Series(["grp1", "grp1", "grp1", "grp1", "grp2", "grp2", "grp2", "grp2"], index=[1, 2, 3, 4, 5, 6, 7, 8])


@pytest.fixture(scope="function")
def XGBoost_classifier(random_state):
    model = XGBClassifier(n_estimators=200, max_depth=3, random_state=random_state)
    return model


@pytest.fixture(scope="function")
def XGBoost_regressor(random_state):
    model = XGBRegressor(n_estimators=200, max_depth=3, random_state=random_state)
    return model


def test_shap_rfe_regressor(XGBoost_regressor, random_state):
    diabetes = load_diabetes()
    X = pd.DataFrame(diabetes.data, columns=diabetes.feature_names)
    y = diabetes.target

    shap_elimination = ShapRFECV(XGBoost_regressor, step=0.8, cv=2, scoring="r2", n_jobs=4, random_state=random_state)
    report = shap_elimination.fit_compute(X, y)

    assert report.shape[0] == 3
    assert shap_elimination.get_reduced_features_set(1) == ["bmi"]

    _ = shap_elimination.plot(show=False)


def test_shap_rfe_randomized_search(X, y, randomized_search_decision_tree_classifier, random_state):
    search = randomized_search_decision_tree_classifier
    shap_elimination = ShapRFECV(search, step=0.8, cv=2, scoring="roc_auc", n_jobs=4, random_state=random_state)
    report = shap_elimination.fit_compute(X, y)

    assert report.shape[0] == 2
    assert shap_elimination.get_reduced_features_set(1) == ["col_3"]

    _ = shap_elimination.plot(show=False)


def test_shap_rfe_multi_class(X, y, decision_tree_classifier, random_state):
    shap_elimination = ShapRFECV(
        decision_tree_classifier,
        cv=2,
        scoring="roc_auc_ovr",
        random_state=random_state,
    )

    report = shap_elimination.fit_compute(X, y, approximate=False, check_additivity=False)

    assert report.shape[0] == 3
    assert shap_elimination.get_reduced_features_set(1) == ["col_3"]


def test_shap_rfe(X, y, sample_weight, decision_tree_classifier, random_state):
    shap_elimination = ShapRFECV(
        decision_tree_classifier,
        random_state=random_state,
        step=1,
        cv=2,
        scoring="roc_auc",
        n_jobs=4,
    )
    report = shap_elimination.fit_compute(X, y, sample_weight=sample_weight, approximate=True, check_additivity=False)

    assert report.shape[0] == 3
    assert shap_elimination.get_reduced_features_set(1) == ["col_3"]


def test_shap_rfe_group_cv(X, y, groups, sample_weight, decision_tree_classifier, random_state):
    cv = StratifiedGroupKFold(n_splits=2, shuffle=True, random_state=random_state)
    shap_elimination = ShapRFECV(
        decision_tree_classifier,
        random_state=random_state,
        step=1,
        cv=cv,
        scoring="roc_auc",
        n_jobs=4,
    )
    report = shap_elimination.fit_compute(
        X, y, groups=groups, sample_weight=sample_weight, approximate=True, check_additivity=False
    )

    assert report.shape[0] == 3
    assert shap_elimination.get_reduced_features_set(1) == ["col_3"]


def test_shap_pipeline_error(X, y, decision_tree_classifier, random_state):
    model = Pipeline(
        [
            ("scaler", StandardScaler()),
            ("dt", decision_tree_classifier),
        ]
    )
    with pytest.raises(TypeError):
        shap_elimination = ShapRFECV(
            model,
            random_state=random_state,
            step=1,
            cv=2,
            scoring="roc_auc",
            n_jobs=4,
        )
        shap_elimination = shap_elimination.fit(X, y, approximate=True, check_additivity=False)


def test_shap_rfe_linear_model(X, y, random_state):
    model = LogisticRegression(C=1, random_state=random_state)
    shap_elimination = ShapRFECV(model, random_state=random_state, step=1, cv=2, scoring="roc_auc", n_jobs=4)
    report = shap_elimination.fit_compute(X, y)

    assert report.shape[0] == 3
    assert shap_elimination.get_reduced_features_set(1) == ["col_3"]


def test_shap_rfe_svm(X, y, random_state):
    model = SVC(C=1, kernel="linear", probability=True, random_state=random_state)
    shap_elimination = ShapRFECV(model, random_state=random_state, step=1, cv=2, scoring="roc_auc", n_jobs=4)
    shap_elimination = shap_elimination.fit(X, y)
    report = shap_elimination.compute()

    assert report.shape[0] == 3
    assert shap_elimination.get_reduced_features_set(1) == ["col_3"]


def test_shap_rfe_cols_to_keep(X, y, decision_tree_classifier, random_state):
    shap_elimination = ShapRFECV(
        decision_tree_classifier,
        random_state=random_state,
        step=2,
        cv=2,
        scoring="roc_auc",
        n_jobs=4,
        min_features_to_select=1,
    )
    report = shap_elimination.fit_compute(X, y, columns_to_keep=["col_2", "col_3"])

    assert report.shape[0] == 2
    reduced_feature_set = set(shap_elimination.get_reduced_features_set(num_features=2))
    assert reduced_feature_set == {"col_2", "col_3"}


def test_shap_rfe_randomized_search_cols_to_keep(X, y, randomized_search_decision_tree_classifier, random_state):
    search = randomized_search_decision_tree_classifier
    shap_elimination = ShapRFECV(search, step=0.8, cv=2, scoring="roc_auc", n_jobs=4, random_state=random_state)
    report = shap_elimination.fit_compute(X, y, columns_to_keep=["col_2", "col_3"])

    assert report.shape[0] == 2
    reduced_feature_set = set(shap_elimination.get_reduced_features_set(num_features=2))
    assert reduced_feature_set == {"col_2", "col_3"}


def test_calculate_number_of_features_to_remove():
    assert 3 == ShapRFECV._calculate_number_of_features_to_remove(
        current_num_of_features=10, num_features_to_remove=3, min_num_features_to_keep=5
    )
    assert 3 == ShapRFECV._calculate_number_of_features_to_remove(
        current_num_of_features=8, num_features_to_remove=5, min_num_features_to_keep=5
    )
    assert 0 == ShapRFECV._calculate_number_of_features_to_remove(
        current_num_of_features=5, num_features_to_remove=1, min_num_features_to_keep=5
    )
    assert 4 == ShapRFECV._calculate_number_of_features_to_remove(
        current_num_of_features=5, num_features_to_remove=7, min_num_features_to_keep=1
    )


def test_shap_automatic_num_feature_selection(decision_tree_classifier, random_state):
    X = pd.DataFrame(
        {
            "col_1": [1, 0, 1, 0, 1, 0, 1, 0],
            "col_2": [0, 0, 0, 0, 0, 1, 1, 1],
            "col_3": [1, 1, 1, 0, 0, 0, 0, 0],
        }
    )
    y = pd.Series([0, 0, 0, 0, 1, 1, 1, 1])

    shap_elimination = ShapRFECV(
        decision_tree_classifier,
        random_state=random_state,
        step=1,
        cv=2,
        scoring="roc_auc",
        n_jobs=1,
    )
    _ = shap_elimination.fit_compute(X, y, approximate=True, check_additivity=False)

    best_features = shap_elimination.get_reduced_features_set(num_features="best")
    best_coherent_features = shap_elimination.get_reduced_features_set(
        num_features="best_coherent",
    )
    best_parsimonious_features = shap_elimination.get_reduced_features_set(num_features="best_parsimonious")

    assert best_features == ["col_2"]
    assert best_coherent_features == ["col_1", "col_2", "col_3"]
    assert best_parsimonious_features == ["col_2"]


def test_get_feature_shap_values_per_fold(X, y, decision_tree_classifier, random_state):
    shap_elimination = ShapRFECV(decision_tree_classifier, scoring="roc_auc", random_state=random_state)
    (
        shap_values,
        train_score,
        test_score,
    ) = shap_elimination._get_feature_shap_values_per_fold(
        X,
        y,
        decision_tree_classifier,
        train_index=[2, 3, 4, 5, 6, 7],
        val_index=[0, 1],
    )
    assert test_score == 1
    assert train_score > 0.9
    assert shap_values.shape == (2, 3)


def test_shap_rfe_same_features_are_kept_after_each_run(random_state_1234):
    """
    Test a use case which appears to be flickering with Probatus 1.8.9 and lower.

    Expected result: every run the same outcome.
    Probatus <= 1.8.9: A different order every time.
    """
    feature_names = [(f"f{num}") for num in range(1, 21)]

    # Code from tutorial on probatus documentation
    X, y = make_classification(
        n_samples=100,
        class_sep=0.05,
        n_informative=6,
        n_features=20,
        random_state=random_state_1234,
        n_redundant=10,
        n_clusters_per_class=1,
    )
    X = pd.DataFrame(X, columns=feature_names)

    random_forest = RandomForestClassifier(
        random_state=random_state_1234,
        n_estimators=70,
        max_features="log2",
        criterion="entropy",
        class_weight="balanced",
    )

    shap_elimination = ShapRFECV(
        random_forest,
        step=0.2,
        cv=5,
        scoring="f1_macro",
        n_jobs=1,
        random_state=random_state_1234,
    )

    report = shap_elimination.fit_compute(X, y, check_additivity=True)
    # Return the set of features with the best validation accuracy

    kept_features = list(report.iloc[[report["val_metric_mean"].idxmax() - 1]]["features_set"].to_list()[0])

    # Results from the first run
    assert [
        "f1",
        "f2",
        "f3",
        "f5",
        "f6",
        "f10",
        "f11",
        "f12",
        "f13",
        "f14",
        "f15",
        "f16",
        "f17",
        "f18",
        "f19",
        "f20",
    ] == kept_features


def test_shap_rfe_penalty_factor(X, y, decision_tree_classifier, random_state):
    shap_elimination = ShapRFECV(
        decision_tree_classifier,
        random_state=random_state,
        step=1,
        cv=2,
        scoring="roc_auc",
        n_jobs=1,
    )
    report = shap_elimination.fit_compute(
        X, y, shap_variance_penalty_factor=1.0, approximate=True, check_additivity=False
    )

    assert report.shape[0] == 3
    assert shap_elimination.get_reduced_features_set(1) == ["col_1"]


def test_complex_dataset(complex_data, complex_lightgbm, random_state_1):
    X, y = complex_data

    param_grid = {
        "n_estimators": [5, 7, 10],
        "num_leaves": [3, 5, 7, 10],
    }
    search = RandomizedSearchCV(complex_lightgbm, param_grid, n_iter=1, random_state=random_state_1)

    shap_elimination = ShapRFECV(
        model=search, step=1, cv=10, scoring="roc_auc", n_jobs=3, verbose=1, random_state=random_state_1
    )

    report = shap_elimination.fit_compute(X, y)

    assert report.shape[0] == X.shape[1]


def test_shap_rfe_early_stopping_lightGBM(complex_data, random_state):
    model = LGBMClassifier(n_estimators=200, max_depth=3, random_state=random_state)
    X, y = complex_data

    shap_elimination = EarlyStoppingShapRFECV(
        model,
        random_state=random_state,
        step=1,
        cv=10,
        scoring="roc_auc",
        n_jobs=4,
        early_stopping_rounds=5,
        eval_metric="auc",
    )
    report = shap_elimination.fit_compute(X, y, approximate=False, check_additivity=False)

    assert report.shape[0] == 5
    assert shap_elimination.get_reduced_features_set(1) == ["f5"]


def test_shap_rfe_early_stopping_XGBoost(XGBoost_classifier, complex_data, random_state):
    X, y = complex_data
    X["f1_categorical"] = X["f1_categorical"].astype(float)

    shap_elimination = EarlyStoppingShapRFECV(
        XGBoost_classifier,
        random_state=random_state,
        step=1,
        cv=10,
        scoring="roc_auc",
        n_jobs=4,
        early_stopping_rounds=5,
        eval_metric="auc",
    )
    report = shap_elimination.fit_compute(X, y, approximate=False, check_additivity=False)

    assert report.shape[0] == 5
    assert shap_elimination.get_reduced_features_set(1) == ["f4"]


#
#
def test_shap_rfe_early_stopping_CatBoost(complex_data_with_categorical, catboost_classifier, random_state):
    X, y = complex_data_with_categorical

    shap_elimination = EarlyStoppingShapRFECV(
        catboost_classifier,
        random_state=random_state,
        step=1,
        cv=10,
        scoring="roc_auc",
        n_jobs=4,
        early_stopping_rounds=5,
        eval_metric="auc",
    )
    report = shap_elimination.fit_compute(X, y, approximate=False, check_additivity=False)

    assert report.shape[0] == 5
    assert shap_elimination.get_reduced_features_set(1)[0] in ["f4", "f5"]


def test_shap_rfe_randomized_search_early_stopping_lightGBM(complex_data, random_state):
    model = LGBMClassifier(n_estimators=200, random_state=random_state)
    X, y = complex_data

    param_grid = {
        "max_depth": [3, 4, 5],
    }
    search = RandomizedSearchCV(model, param_grid, cv=2, n_iter=2, random_state=random_state)
    shap_elimination = EarlyStoppingShapRFECV(
        search,
        step=1,
        cv=10,
        scoring="roc_auc",
        early_stopping_rounds=5,
        eval_metric="auc",
        n_jobs=4,
        verbose=1,
        random_state=random_state,
    )
    report = shap_elimination.fit_compute(X, y)

    assert report.shape[0] == X.shape[1]
    assert shap_elimination.get_reduced_features_set(1) == ["f5"]

    _ = shap_elimination.plot(show=False)


def test_get_feature_shap_values_per_fold_early_stopping_lightGBM(complex_data, random_state):
    model = LGBMClassifier(n_estimators=200, max_depth=3, random_state=random_state)
    X, y = complex_data
    y = preprocess_labels(y, y_name="y", index=X.index)

    shap_elimination = EarlyStoppingShapRFECV(
        model, early_stopping_rounds=5, scoring="roc_auc", random_state=random_state
    )
    (
        shap_values,
        train_score,
        test_score,
    ) = shap_elimination._get_feature_shap_values_per_fold(
        X,
        y,
        model,
        train_index=list(range(5, 50)),
        val_index=[0, 1, 2, 3, 4],
    )
    assert test_score > 0.6
    assert train_score > 0.6
    assert shap_values.shape == (5, 5)


def test_get_feature_shap_values_per_fold_early_stopping_CatBoost(
    complex_data_with_categorical, catboost_classifier, random_state
):
    X, y = complex_data_with_categorical
    y = preprocess_labels(y, y_name="y", index=X.index)

    shap_elimination = EarlyStoppingShapRFECV(
        catboost_classifier, early_stopping_rounds=5, scoring="roc_auc", random_state=random_state
    )
    (
        shap_values,
        train_score,
        test_score,
    ) = shap_elimination._get_feature_shap_values_per_fold(
        X,
        y,
        catboost_classifier,
        train_index=list(range(5, 50)),
        val_index=[0, 1, 2, 3, 4],
    )
    assert test_score > 0
    assert train_score > 0.6
    assert shap_values.shape == (5, 5)


def test_get_feature_shap_values_per_fold_early_stopping_XGBoost(XGBoost_classifier, complex_data, random_state):
    X, y = complex_data
    y = preprocess_labels(y, y_name="y", index=X.index)

    shap_elimination = EarlyStoppingShapRFECV(
        XGBoost_classifier, early_stopping_rounds=5, scoring="roc_auc", random_state=random_state
    )
    (
        shap_values,
        train_score,
        test_score,
    ) = shap_elimination._get_feature_shap_values_per_fold(
        X,
        y,
        XGBoost_classifier,
        train_index=list(range(5, 50)),
        val_index=[0, 1, 2, 3, 4],
    )
    assert test_score > 0
    assert train_score > 0.6
    assert shap_values.shape == (5, 5)


def test_EarlyStoppingShapRFECV_no_categorical(complex_data, random_state):
    model = LGBMClassifier(n_estimators=50, max_depth=3, num_leaves=3, random_state=random_state)

    shap_elimination = EarlyStoppingShapRFECV(
        model=model,
        step=0.33,
        cv=5,
        scoring="accuracy",
        eval_metric="logloss",
        early_stopping_rounds=5,
        random_state=random_state,
    )
    X, y = complex_data
    X = X.drop(columns=["f1_categorical"])
    report = shap_elimination.fit_compute(X, y, feature_perturbation="tree_path_dependent")

    assert report.shape[0] == X.shape[1]
    assert shap_elimination.get_reduced_features_set(1) == ["f5"]

    _ = shap_elimination.plot(show=False)


def test_LightGBM_stratified_kfold(random_state):
    """
    Test added to check for https://github.com/ing-bank/probatus/issues/170.
    """
    X = pd.DataFrame(
        [
            [1, 2, 3, 4, 5, 101, 102, 103, 104, 105],
            [-1, -2, 2, -5, -7, 1, 2, 5, -1, 3],
            ["a", "b"] * 5,  # noisy categorical will dropped first
        ]
    ).transpose()
    X[2] = X[2].astype("category")
    X[1] = X[1].astype("float")
    X[0] = X[0].astype("float")
    y = [0] * 5 + [1] * 5

    model = LGBMClassifier(random_state=random_state)
    n_iter = 2
    n_folds = 3

    for _ in range(n_iter):
        skf = StratifiedKFold(n_folds, shuffle=True, random_state=random_state)
        shap_elimination = EarlyStoppingShapRFECV(
            model=model,
            step=1 / (n_iter + 1),
            cv=skf,
            scoring="accuracy",
            eval_metric="logloss",
            early_stopping_rounds=5,
            random_state=random_state,
        )
        report = shap_elimination.fit_compute(X, y, feature_perturbation="tree_path_dependent")

    assert report.shape[0] == X.shape[1]

    shap_elimination.plot(show=False)



from sklearn.metrics import get_scorer


def get_single_scorer(scoring):
    """
    Returns Scorer, based on provided input in scoring argument.

    Args:
        scoring (string or probatus.utils.Scorer, optional):
            Metric for which the model performance is calculated. It can be either a metric name aligned with
            predefined classification scorers names in sklearn
            ([link](https://scikit-learn.org/stable/modules/model_evaluation.html)).
            Another option is using probatus.utils.Scorer to define a custom metric.

    Returns:
        (probatus.utils.Scorer):
            Scorer that can be used for scoring models
    """
    if isinstance(scoring, str):
        return Scorer(scoring)
    elif isinstance(scoring, Scorer):
        return scoring
    else:
        raise (ValueError("The scoring should contain either strings or probatus.utils.Scorer class"))


class Scorer:
    """
    Scores a given machine learning model based on the provided metric name and optionally a custom scoring function.

    Examples:

    ```python
    from probatus.utils import Scorer
    from sklearn.metrics import make_scorer
    from sklearn.datasets import make_classification
    from sklearn.model_selection import train_test_split
    from sklearn.ensemble import RandomForestClassifier
    import pandas as pd

    # Make ROC AUC scorer
    scorer1 = Scorer('roc_auc')

    # Make custom scorer with following function:
    def custom_metric(y_true, y_pred):
         return (y_true == y_pred).sum()
    scorer2 = Scorer('custom_metric', custom_scorer=make_scorer(custom_metric))

    # Prepare two samples
    feature_names = ['f1', 'f2', 'f3', 'f4']
    X, y = make_classification(n_samples=1000, n_features=4, random_state=0)
    X = pd.DataFrame(X, columns=feature_names)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)

    # Prepare and fit model. Remember about class_weight="balanced" or an equivalent.
    model = RandomForestClassifier(class_weight='balanced', n_estimators = 100, max_depth=2, random_state=0)
    model = model.fit(X_train, y_train)

    # Score model
    score_test_scorer1 = scorer1.score(model, X_test, y_test)
    score_test_scorer2 = scorer2.score(model, X_test, y_test)

    print(f'Test ROC AUC is {score_test_scorer1}, Test {scorer2.metric_name} is {score_test_scorer2}')
    ```
    """

    def __init__(self, metric_name, custom_scorer=None):
        """
        Initializes the class.

        Args:
            metric_name (str): Name of the metric used to evaluate the model.
                If the custom_scorer is not passed, the
                metric name needs to be aligned with classification scorers names in sklearn
                ([link](https://scikit-learn.org/stable/modules/model_evaluation.html)).
            custom_scorer (sklearn.metrics Scorer callable, optional): Callable
                that can score samples.
        """
        self.metric_name = metric_name
        if custom_scorer is not None:
            self.scorer = custom_scorer
        else:
            self.scorer = get_scorer(self.metric_name)

    def score(self, model, X, y):
        """
        Scores the samples model based on the provided metric name.

        Args
            model (model object):
                Model to be scored.

            X (array-like of shape (n_samples,n_features)):
                Samples on which the model is scored.

            y (array-like of shape (n_samples,)):
                Labels on which the model is scored.

        Returns:
            (float):
                Score returned by the model
        """
        return self.scorer(model, X, y)



import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytest
from pandas.api.types import is_numeric_dtype

from probatus.sample_similarity import BaseResemblanceModel, PermutationImportanceResemblance, SHAPImportanceResemblance

# Turn off interactive mode in plots
plt.ioff()
matplotlib.use("Agg")


@pytest.fixture(scope="function")
def X1():
    return pd.DataFrame({"col_1": [1, 1, 1, 1], "col_2": [0, 0, 0, 0], "col_3": [0, 0, 0, 0]}, index=[1, 2, 3, 4])


@pytest.fixture(scope="function")
def X2():
    return pd.DataFrame({"col_1": [0, 0, 0, 0], "col_2": [0, 0, 0, 0], "col_3": [0, 0, 0, 0]}, index=[1, 2, 3, 4])


def test_base_class(X1, X2, decision_tree_classifier, random_state):
    rm = BaseResemblanceModel(decision_tree_classifier, test_prc=0.5, n_jobs=1, random_state=random_state)

    actual_report, train_score, test_score = rm.fit_compute(X1, X2, return_scores=True)

    assert train_score == 1
    assert test_score == 1
    assert actual_report is None

    # Check data splits if correct
    actual_X_train, actual_X_test, actual_y_train, actual_y_test = rm.get_data_splits()

    assert actual_X_train.shape == (4, 3)
    assert actual_X_test.shape == (4, 3)
    assert len(actual_y_train) == 4
    assert len(actual_y_test) == 4

    # Check if stratified
    assert np.sum(actual_y_train) == 2
    assert np.sum(actual_y_test) == 2

    # Check if index is correct
    assert len(rm.X.index.unique()) == 8
    assert list(rm.X.index) == list(rm.y.index)

    with pytest.raises(NotImplementedError) as _:
        rm.plot()


def test_base_class_lin_models(X1, X2, logistic_regression, random_state):
    # Test class BaseResemblanceModel for linear models.
    rm = BaseResemblanceModel(logistic_regression, test_prc=0.5, n_jobs=1, random_state=random_state)

    actual_report, train_score, test_score = rm.fit_compute(X1, X2, return_scores=True)

    assert train_score == 1
    assert test_score == 1
    assert actual_report is None

    # Check data splits if correct
    actual_X_train, actual_X_test, actual_y_train, actual_y_test = rm.get_data_splits()

    assert actual_X_train.shape == (4, 3)
    assert actual_X_test.shape == (4, 3)
    assert len(actual_y_train) == 4
    assert len(actual_y_test) == 4

    # Check if stratified
    assert np.sum(actual_y_train) == 2
    assert np.sum(actual_y_test) == 2

    # Check if index is correct
    assert len(rm.X.index.unique()) == 8
    assert list(rm.X.index) == list(rm.y.index)

    with pytest.raises(NotImplementedError) as _:
        rm.plot()


def test_shap_resemblance_class(X1, X2, decision_tree_classifier, random_state):
    rm = SHAPImportanceResemblance(decision_tree_classifier, test_prc=0.5, n_jobs=1, random_state=random_state)

    actual_report, train_score, test_score = rm.fit_compute(X1, X2, return_scores=True)

    assert train_score == 1
    assert test_score == 1

    # Check report shape
    assert actual_report.shape == (3, 2)
    # Check if it is sorted by importance
    assert actual_report.iloc[0].name == "col_1"
    # Check report values
    assert actual_report.loc["col_1"]["mean_abs_shap_value"] > 0
    assert actual_report.loc["col_1"]["mean_shap_value"] < 0
    assert actual_report.loc["col_2"]["mean_abs_shap_value"] == 0
    assert actual_report.loc["col_2"]["mean_shap_value"] == 0
    assert actual_report.loc["col_3"]["mean_abs_shap_value"] == 0
    assert actual_report.loc["col_3"]["mean_shap_value"] == 0

    actual_shap_values_test = rm.get_shap_values()
    assert actual_shap_values_test.shape == (4, 3)

    # Run plots
    rm.plot(plot_type="bar")
    rm.plot(plot_type="dot")


def test_shap_resemblance_class_lin_models(X1, X2, logistic_regression, random_state):
    # Test SHAP Resemblance Model for linear models.
    rm = SHAPImportanceResemblance(logistic_regression, test_prc=0.5, n_jobs=1, random_state=random_state)

    actual_report, train_score, test_score = rm.fit_compute(
        X1, X2, return_scores=True, approximate=True, check_additivity=False
    )

    assert train_score == 1
    assert test_score == 1

    # Check report shape
    assert actual_report.shape == (3, 2)
    # Check if it is sorted by importance
    assert actual_report.iloc[0].name == "col_1"
    # Check report values
    assert actual_report.loc["col_1"]["mean_abs_shap_value"] > 0
    assert actual_report.loc["col_1"]["mean_shap_value"] < 0
    assert actual_report.loc["col_2"]["mean_abs_shap_value"] == 0
    assert actual_report.loc["col_2"]["mean_shap_value"] == 0
    assert actual_report.loc["col_3"]["mean_abs_shap_value"] == 0
    assert actual_report.loc["col_3"]["mean_shap_value"] == 0

    actual_shap_values_test = rm.get_shap_values()
    assert actual_shap_values_test.shape == (4, 3)

    # Run plots
    rm.plot(plot_type="bar")
    rm.plot(plot_type="dot")


def test_shap_resemblance_class2(complex_data_with_categorical, complex_lightgbm, random_state):
    X1, _ = complex_data_with_categorical
    X2 = X1.copy()
    X2["f4"] = X2["f4"] + 100

    rm = SHAPImportanceResemblance(
        complex_lightgbm, scoring="accuracy", test_prc=0.5, n_jobs=1, random_state=random_state
    )

    actual_report, train_score, test_score = rm.fit_compute(X1, X2, return_scores=True, class_names=["a", "b"])

    # Check if the X and y within the rm have correct types
    assert rm.X["f1_categorical"].dtype.name == "category"
    for num_column in ["f2_missing", "f3_static", "f4", "f5"]:
        assert is_numeric_dtype(rm.X[num_column])

    assert train_score == pytest.approx(1, 0.05)
    assert test_score == pytest.approx(1, 0.05)

    # Check report shape
    assert actual_report.shape == (5, 2)
    # Check if it is sorted by importance
    assert actual_report.iloc[0].name == "f4"

    # Check report values
    assert actual_report.loc["f4"]["mean_abs_shap_value"] > 0

    actual_shap_values_test = rm.get_shap_values()
    # 50 test samples and 5 features
    assert actual_shap_values_test.shape == (X1.shape[0], X1.shape[1])

    # Run plots
    rm.plot(plot_type="bar", show=True)
    rm.plot(plot_type="dot", show=False)


def test_permutation_resemblance_class(X1, X2, decision_tree_classifier, random_state):
    rm = PermutationImportanceResemblance(
        decision_tree_classifier, test_prc=0.5, n_jobs=1, random_state=random_state, iterations=20
    )

    actual_report, train_score, test_score = rm.fit_compute(X1, X2, return_scores=True)

    assert train_score == 1
    assert test_score == 1

    # Check report shape
    assert actual_report.shape == (3, 2)
    # Check if it is sorted by importance
    assert actual_report.iloc[0].name == "col_1"
    # Check report values
    assert actual_report.loc["col_1"]["mean_importance"] > 0
    assert actual_report.loc["col_1"]["std_importance"] > 0
    assert actual_report.loc["col_2"]["mean_importance"] == 0
    assert actual_report.loc["col_2"]["std_importance"] == 0
    assert actual_report.loc["col_3"]["mean_importance"] == 0
    assert actual_report.loc["col_3"]["std_importance"] == 0

    rm.plot(figsize=(10, 10))
    # Check plot size
    fig = plt.gcf()
    size = fig.get_size_inches()
    assert size[0] == 10 and size[1] == 10


def test_base_class_same_data(X1, decision_tree_classifier, random_state):
    rm = BaseResemblanceModel(decision_tree_classifier, test_prc=0.5, n_jobs=1, random_state=random_state)

    actual_report, train_score, test_score = rm.fit_compute(X1, X1, return_scores=True)

    assert train_score == 0.5
    assert test_score == 0.5
    assert actual_report is None



from probatus.interpret import ShapModelInterpreter
import pytest
from probatus.utils import NotFittedError


def test_fitted_exception(fitted_tree, X_train, y_train, X_test, y_test, random_state):
    class_names = ["neg", "pos"]

    shap_interpret = ShapModelInterpreter(fitted_tree, random_state=random_state)

    # Before fit it should raise an exception
    with pytest.raises(NotFittedError) as _:
        shap_interpret._check_if_fitted()

    shap_interpret.fit(X_train, X_test, y_train, y_test, class_names=class_names)

    # Check parameters
    assert shap_interpret.fitted
    shap_interpret._check_if_fitted


@pytest.mark.xfail
def test_fitted_exception_is_raised(fitted_tree, random_state):
    shap_interpret = ShapModelInterpreter(fitted_tree, random_state=random_state)

    shap_interpret._check_if_fitted



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.preprocessing import KBinsDiscretizer

from probatus.utils import BaseFitComputePlotClass, preprocess_data, preprocess_labels, shap_to_df


class DependencePlotter(BaseFitComputePlotClass):
    """
    Plotter used to plot SHAP dependence plot together with the target rates.

    Currently it supports tree-based and linear models.

    Args:
        model: classifier for which interpretation is done.

    Example:
    ```python
    from sklearn.datasets import make_classification
    from sklearn.ensemble import RandomForestClassifier
    from probatus.interpret import DependencePlotter

    X, y = make_classification(n_samples=15, n_features=3, n_informative=3, n_redundant=0, random_state=42)
    model = RandomForestClassifier().fit(X, y)
    bdp = DependencePlotter(model)
    shap_values = bdp.fit_compute(X, y)

    bdp.plot(feature=2)
    ```

    <img src="../img/model_interpret_dep.png"/>
    """

    def __init__(self, model, verbose=0, random_state=None):
        """
        Initializes the class.

        Args:
            model (model object):
                regression or classification model or pipeline.

            verbose (int, optional):
                Controls verbosity of the output:

                - 0 - neither prints nor warnings are shown
                - 1 - only most important warnings
                - 2 - shows all prints and all warnings.

            random_state (int, optional):
                Random state set for the nr of samples. If it is None, the results will not be reproducible. For
                reproducible results set it to an integer.
        """
        self.model = model
        self.verbose = verbose
        self.random_state = random_state

    def __repr__(self):
        """
        Represent string method.
        """
        return f"Shap dependence plotter for {self.model.__class__.__name__}"

    def fit(self, X, y, column_names=None, class_names=None, precalc_shap=None, **shap_kwargs):
        """
        Fits the plotter to the model and data by computing the shap values.

        If the shap_values are passed, they do not need to be computed.

        Args:
            X (pd.DataFrame): input variables.

            y (pd.Series): target variable.

            column_names (None, or list of str, optional):
                List of feature names for the dataset. If None, then column names from the X_train dataframe are used.

            class_names (None, or list of str, optional):
                List of class names e.g. ['neg', 'pos']. If none, the default ['Negative Class', 'Positive Class'] are
                used.

            precalc_shap (Optional, None or np.array):
                Precalculated shap values, If provided they don't need to be computed.

            **shap_kwargs:
                keyword arguments passed to
                [shap.Explainer](https://shap.readthedocs.io/en/latest/generated/shap.Explainer.html#shap.Explainer).
                It also enables `approximate` and `check_additivity` parameters, passed while calculating SHAP values.
                The `approximate=True` causes less accurate, but faster SHAP values calculation, while
                `check_additivity=False` disables the additivity check inside SHAP.
        """
        self.X, self.column_names = preprocess_data(X, X_name="X", column_names=column_names, verbose=self.verbose)
        self.y = preprocess_labels(y, y_name="y", index=self.X.index, verbose=self.verbose)

        # Set class names
        self.class_names = class_names
        if self.class_names is None:
            self.class_names = ["Negative Class", "Positive Class"]

        self.shap_vals_df = shap_to_df(
            self.model,
            self.X,
            precalc_shap=precalc_shap,
            verbose=self.verbose,
            random_state=self.random_state,
            **shap_kwargs,
        )

        self.fitted = True
        return self

    def compute(self):
        """
        Computes the report returned to the user, namely the SHAP values generated on the dataset.

        Returns:
            (pd.DataFrame):
                SHAP Values for X.
        """
        self._check_if_fitted()
        return self.shap_vals_df

    def fit_compute(self, X, y, column_names=None, class_names=None, precalc_shap=None, **shap_kwargs):
        """
        Fits the plotter to the model and data by computing the shap values.

        If the shap_values are passed, they do not need to be computed

        Args:
            X (pd.DataFrame):
                Provided dataset.

            y (pd.Series):
                Labels for X.

            column_names (None, or list of str, optional):
                List of feature names for the dataset. If None, then column names from the X_train dataframe are used.

            class_names (None, or list of str, optional):
                List of class names e.g. ['neg', 'pos']. If none, the default ['Negative Class', 'Positive Class'] are
                used.

            precalc_shap (Optional, None or np.array):
                Precalculated shap values, If provided they don't need to be computed.

            **shap_kwargs:
                keyword arguments passed to
                [shap.Explainer](https://shap.readthedocs.io/en/latest/generated/shap.Explainer.html#shap.Explainer).
                It also enables `approximate` and `check_additivity` parameters, passed while calculating SHAP values.
                The `approximate=True` causes less accurate, but faster SHAP values calculation, while
                `check_additivity=False` disables the additivity check inside SHAP.

        Returns:
            (pd.DataFrame):
                SHAP Values for X.
        """
        self.fit(X, y, column_names=column_names, class_names=class_names, precalc_shap=precalc_shap, **shap_kwargs)
        return self.compute()

    def plot(
        self,
        feature,
        figsize=(15, 10),
        bins=10,
        show=True,
        min_q=0,
        max_q=1,
        alpha=1.0,
    ):
        """
        Plots the shap values for data points for a given feature, as well as the target rate and values distribution.

        Args:
            feature (str or int):
                Feature name of the feature to be analyzed.

            figsize ((float, float), optional):
                Tuple specifying size (width, height) of resulting figure in inches.

            bins (int or list[float]):
                Number of bins or boundaries of bins (supplied in list) for target-rate plot.

            show (bool, optional):
                If True, the plots are showed to the user, otherwise they are not shown. Not showing plot can be useful,
                when you want to edit the returned axis, before showing it.

            min_q (float, optional):
                Optional minimum quantile from which to consider values, used for plotting under outliers.

            max_q (float, optional):
                Optional maximum quantile until which data points are considered, used for plotting under outliers.

            alpha (float, optional):
                Optional alpha blending value, between 0 (transparent) and 1 (opaque).

        Returns
            (list(matplotlib.axes)):
                List of axes that include the plots.
        """
        self._check_if_fitted()
        if min_q >= max_q:
            raise ValueError("min_q must be smaller than max_q")
        if feature not in self.X.columns:
            raise ValueError("Feature not recognized")
        if (alpha < 0) or (alpha > 1):
            raise ValueError("alpha must be a float value between 0 and 1")

        self.min_q, self.max_q, self.alpha = min_q, max_q, alpha

        _ = plt.figure(1, figsize=figsize)
        ax1 = plt.subplot2grid((3, 1), (0, 0), rowspan=2)
        ax2 = plt.subplot2grid((3, 1), (2, 0))

        self._dependence_plot(feature=feature, ax=ax1)
        self._target_rate_plot(feature=feature, bins=bins, ax=ax2)

        ax2.set_xlim(ax1.get_xlim())

        if show:
            plt.show()
        else:
            plt.close()

        return [ax1, ax2]

    def _dependence_plot(self, feature, ax=None):
        """
        Plots shap values for data points with respect to specified feature.

        Args:
            feature (str or int):
                Feature for which dependence plot is to be created.

            ax (matplotlib.pyplot.axes, optional):
                Optional axis on which to draw plot.

        Returns:
            (matplotlib.pyplot.axes):
                Axes on which plot is drawn.
        """
        if isinstance(feature, int):
            feature = self.column_names[feature]

        X, y, shap_val = self._get_X_y_shap_with_q_cut(feature=feature)

        ax.scatter(X[y == 0], shap_val[y == 0], label=self.class_names[0], color="lightblue", alpha=self.alpha)

        ax.scatter(X[y == 1], shap_val[y == 1], label=self.class_names[1], color="darkred", alpha=self.alpha)

        ax.set_ylabel("Shap value")
        ax.set_title(f"Dependence plot for {feature} feature")
        ax.legend()

        return ax

    def _target_rate_plot(self, feature, bins=10, ax=None):
        """
        Plots the distributions of the specific features, as well as the target rate as function of the feature.

        Args:
            feature (str or int):
                Feature for which to create target rate plot.

            bins (int or list[float]), optional:
                Number of bins or boundaries of desired bins in list.

            ax (matplotlib.pyplot.axes, optional):
                Optional axis on which to draw plot.

        Returns:
            (list[float], matplotlib.pyplot.axes, float):
                Tuple of boundaries of bins used, axis on which plot is drawn, total ratio of target (positive over
                negative).
        """
        x, y, shap_val = self._get_X_y_shap_with_q_cut(feature=feature)

        # Create bins if not explicitly supplied
        if isinstance(bins, int):
            simple_binner = KBinsDiscretizer(n_bins=bins, encode="ordinal", strategy="uniform").fit(
                np.array(x).reshape(-1, 1)
            )
            bins = simple_binner.bin_edges_[0]
            bins[0], bins[-1] = -np.inf, np.inf

        # Determine bin for datapoints
        bins[-1] = bins[-1] + 1
        indices = np.digitize(x, bins)
        # Create dataframe with binned data
        dfs = pd.DataFrame({feature: x, "y": y, "bin_index": pd.Series(indices, index=x.index)}).groupby(
            "bin_index", as_index=True
        )

        # Extract target ratio and mean feature value
        target_ratio = dfs["y"].mean()
        x_vals = dfs[feature].mean()

        # Transform the first and last bin to work with plt.hist method
        if bins[0] == -np.inf:
            bins[0] = x.min()
        if bins[-1] == np.inf:
            bins[-1] = x.max()

        # Plot target rate
        ax.hist(x, bins=bins, lw=2, alpha=0.4)
        ax.set_ylabel("Counts")
        ax2 = ax.twinx()
        ax2.plot(x_vals, target_ratio, color="red")
        ax2.set_ylabel("Target rate", color="red", fontsize=12)
        ax2.set_xlim(x.min(), x.max())
        ax.set_xlabel(f"{feature} feature values")

        return bins, ax, target_ratio

    def _get_X_y_shap_with_q_cut(self, feature):
        """
        Extracts all X, y pairs and shap values that fall within defined quantiles of the feature.

        Args:
            feature (str): feature to return values for

        Returns:
            x (pd.Series): selected datapoints
            y (pd.Series): target values of selected datapoints
            shap_val (pd.Series): shap values of selected datapoints
        """
        self._check_if_fitted()
        if feature not in self.X.columns:
            raise ValueError("Feature not found in data")

        # Prepare arrays
        x = self.X[feature]
        y = self.y
        shap_val = self.shap_vals_df[feature]

        # Determine quantile ranges
        x_min = x.quantile(self.min_q)
        x_max = x.quantile(self.max_q)

        # Create filter
        filter = (x >= x_min) & (x <= x_max)

        # Filter and return terms
        return x[filter], y[filter], shap_val[filter]
